"""
Module for Terraform Cloud API Endpoint: Admin Runs.
"""

from .endpoint import TFCEndpoint

class TFCAdminRuns(TFCEndpoint):
    """
    `Admin Runs API Docs \
        <https://www.terraform.io/docs/cloud/api/admin/runs.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/admin/runs"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return True

    def list(self, query=None, filters=None, page=None, page_size=None):
        """
        ``GET /admin/runs``

        `Admin List Runs API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/runs.html#list-all-runs>`_
        """
        return self._list(self._endpoint_base_url, \
            query=query, filters=filters, page=page, page_size=page_size)

    def force_cancel(self, run_id, data=None):
        """
        ``POST /admin/runs/:id/actions/force-cancel``

        `Admin Force Cancel Run API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/runs.html#force-a-run-into-the-quot-cancelled-quot-state>`_
        """
        url = f"{self._endpoint_base_url}/{run_id}/actions/force-cancel"
        return self._post(url, data=data)



"""
Module for testing the Terraform Cloud API Endpoint: Notification Configurations.
"""

from .base import TestTFCBaseTestCase


class TestTFCNotificationConfigurations(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Notification
    Configurations.
    """

    _unittest_name = "not-cng"
    _endpoint_being_tested = "notification_configs"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        # Store the OAuth client ID to remove it at the end.
        oauth_client = self._api.oauth_clients.create(self._get_oauth_client_create_payload())
        self._oauth_client_id = oauth_client["data"]["id"]
        oauth_token_id = oauth_client["data"]["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a workspace using that token ID, save the workspace ID
        ws_payload = self._get_ws_with_vcs_create_payload(oauth_token_id)
        workspace = self._api.workspaces.create(ws_payload)["data"]
        self._ws_id = workspace["id"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_notifications_configurations(self):
        """
        Test the Notification Configurations API endpoints.
        """

        # Show that there are no configuration notifications for the workspace
        all_noti_configs = self._api.notification_configs.list(self._ws_id)["data"]
        self.assertEqual(len(all_noti_configs), 0)

        # Add one notification configuration
        not_cnf_name = self._unittest_random_name()
        payload = {
            "data": {
                "type": "notification-configurations",
                "attributes": {
                    "destination-type": "generic",
                    "enabled": True,
                    "name": not_cnf_name,
                    "url": "https://httpstat.us/200",
                    "triggers": [
                        "run:applying"
                    ]
                }
            }
        }
        created_noti_config = self._api.notification_configs.create(self._ws_id, payload)["data"]
        created_noti_config_id = created_noti_config["id"]

        # Check that the notification config we added is now there
        all_noti_configs = self._api.notification_configs.list(self._ws_id)["data"]
        found_noti_conf = False
        for noti_conf in all_noti_configs:
            if noti_conf["id"] == created_noti_config_id:
                found_noti_conf = True
                break
        self.assertTrue(found_noti_conf)

        # Show the notification configuration we just created, compare the IDs
        shown_noti_config = self._api.notification_configs.show(created_noti_config_id)["data"]
        shown_noti_config_id = shown_noti_config["id"]
        self.assertEqual(shown_noti_config_id, created_noti_config_id)

        # Update the  name of the notification configuration, and check that it worked
        name_to_update_to = "foobar"
        update_payload = {
            "data": {
                "id": created_noti_config_id,
                "type": "notification-configurations",
                "attributes": {
                    "name": name_to_update_to
                }
            }
        }
        updated_noti_config = \
            self._api.notification_configs.update(created_noti_config_id, update_payload)["data"]
        updated_noti_name = updated_noti_config["attributes"]["name"]
        self.assertEqual(updated_noti_name, name_to_update_to)

        # Check that we can verify the notification configuration endpoint
        verified_noti_config = self._api.notification_configs.verify(created_noti_config_id)["data"]
        verified_noti_config_name = verified_noti_config["attributes"]["name"]
        self.assertEqual(name_to_update_to, verified_noti_config_name)

        # Destroy the notification configuraiton, and show the workspace has zero again
        self._api.notification_configs.destroy(created_noti_config_id)
        all_noti_configs = self._api.notification_configs.list(self._ws_id)["data"]
        found_noti_conf = False
        for noti_conf in all_noti_configs:
            if noti_conf["id"] == created_noti_config_id:
                found_noti_conf = True
                break
        self.assertFalse(found_noti_conf)



"""
Module for Terraform Cloud API Endpoint: OAuth Tokens.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCOAuthTokens(TFCEndpoint):
    """
    `OAuth Tokens API Docs \
        <https://www.terraform.io/docs/cloud/api/oauth-tokens.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._oauth_clients_api_v2_base_url = f"{self._api_v2_base_url}/oauth-clients"
        self._oauth_tokens_api_v2_base_url = f"{self._api_v2_base_url}/oauth-tokens"

    def required_entitlements(self):
        return [Entitlements.VCS_INTEGRATIONS]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def list(self, oauth_client_id):
        """
        ``GET /oauth-clients/:oauth_client_id/oauth-tokens``

        `OAuth Tokens List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/oauth-tokens.html#list-oauth-tokens>`_
        """
        url = f"{self._oauth_clients_api_v2_base_url}/{oauth_client_id}/oauth-tokens"
        return self._list(url)

    def show(self, token_id):
        """
        ``GET /oauth-tokens/:id``

        `OAuth Tokens Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/oauth-tokens.html#show-an-oauth-token>`_
        """
        url = f"{self._oauth_tokens_api_v2_base_url}/{token_id}"
        return self._show(url)

    def update(self, token_id, payload):
        """
        ``PATCH /oauth-tokens/:id``

        `OAuth Tokens Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/oauth-tokens.html#update-an-oauth-token>`_
        """
        url = f"{self._oauth_tokens_api_v2_base_url}/{token_id}"
        return self._update(url, payload)

    def destroy(self, token_id):
        """
        ``DELETE /oauth-tokens/:id``

        `OAuth Tokens Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/oauth-tokens.html#destroy-an-oauth-token>`_
        """
        url = f"{self._oauth_tokens_api_v2_base_url}/{token_id}"
        return self._destroy(url)



"""
Module for testing the Terraform Cloud API Endpoint: Applies.
"""

import time

from .base import TestTFCBaseTestCase


class TestTFCApplies(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Applies.
    """

    _unittest_name = "applies"
    _endpoint_being_tested = "applies"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        oauth_client = self._api.oauth_clients.create(self._get_oauth_client_create_payload())
        self._oauth_client_id = oauth_client["data"]["id"]
        oauth_token_id = oauth_client["data"]["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a workspace using that token ID, save the workspace ID
        ws_payload = self._get_ws_with_vcs_create_payload(oauth_token_id)
        workspace = self._api.workspaces.create(ws_payload)["data"]
        self._ws_id = workspace["id"]
        self._temp_org_name = self._random_name()

        # Configure the required variables on the workspace
        variable_payloads = [
            self._get_variable_create_payload(
                "email", self._test_email, self._ws_id),
            self._get_variable_create_payload(
                "org_name", self._temp_org_name, self._ws_id),
            self._get_variable_create_payload(
                "TFE_TOKEN", self._test_api_token, self._ws_id, category="env", sensitive=True)
        ]
        for payload in variable_payloads:
            self._api.vars.create(payload)

        # Sleep for 1 second to give the WS time to create
        time.sleep(1)

        # Start the run, store the run ID
        create_run_payload = self._get_run_create_payload(self._ws_id)
        self._run = self._api.runs.create(create_run_payload)["data"]
        self._run_id = self._run["id"]


    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)
        self._api.orgs.destroy(self._temp_org_name)

    def test_apply(self):
        """
        Test the Applies API endpoints.
        """

        # Timeout if the plan doesn't reach confirmable, this can happen
        # if the run is queued.
        self._created_run_timeout(self._run_id)

        # Apply the plan
        apply_payload = {
            "comment": "foo"
        }
        self._api.runs.apply(self._run_id, apply_payload)
        applied_run = self._api.runs.show(self._run_id)["data"]

        self._logger.debug("Waiting for apply to kick off...")
        while "applying-at" not in applied_run["attributes"]["status-timestamps"]:
            applied_run = self._api.runs.show(self._run_id)["data"]
            time.sleep(1)
        self._logger.debug("Apply kicked off.")
        self.assertIsNotNone(applied_run["attributes"]["status-timestamps"]["applying-at"])

        # Show the apply, confirm it returns the same value as the run attributes
        apply_id = applied_run["relationships"]["apply"]["data"]["id"]
        shown_apply = self._api.applies.show(apply_id)["data"]
        shown_apply_id = shown_apply["id"]
        self.assertEqual(apply_id, shown_apply_id)

        # Wait for the apply to finish successfully
        self._logger.debug("Waiting for apply to finish...")
        shown_apply = self._api.applies.show(apply_id)["data"]
        while shown_apply["attributes"]["status"] != "finished":
            shown_apply = self._api.applies.show(apply_id)["data"]
            time.sleep(1)
        self.assertEqual("finished", shown_apply["attributes"]["status"])
        self._logger.debug("Apply finished.")



"""
Module for testing the Terraform Cloud API Endpoint: Team Tokens.
"""

from .base import TestTFCBaseTestCase


class TestTFCTeamTokens(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Team Tokens.
    """

    _unittest_name = "team-tok"
    _endpoint_being_tested = "team_tokens"

    def setUp(self):
        self._team = self._api.teams.create(
            self._get_team_create_payload())["data"]
        self._team_id = self._team["id"]

    def tearDown(self):
        self._api.teams.destroy(self._team_id)

    def test_team_tokens(self):
        """
        Test the Team Tokens API endpoints.
        """

        # Create a test token and make sure we get an ID back
        created_token = self._api.team_tokens.create(self._team_id)["data"]
        self.assertIsNotNone(created_token["id"])

        # Then destroy it. There is no lookup so there's nothing to test here.
        self._api.team_tokens.destroy(self._team_id)



"""
Module for testing the Terraform Cloud API Endpoint: Workspaces.
"""

from .base import TestTFCBaseTestCase


class TestTFCWorkspaces(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Workspaces.
    """

    _unittest_name = "ws"
    _endpoint_being_tested = "workspaces"

    def setUp(self):
        # Add an SSH Key to TFC
        created_key = self._api.ssh_keys.create(self._get_ssh_key_create_payload())["data"]
        self._created_key_id = created_key["id"]

    def tearDown(self):
        self._api.ssh_keys.destroy(self._created_key_id)

    def test_workspaces(self):
        """
        Test the Workspaces API endpoints.
        """

        # Get the number of existing workspaces, then create one and compare them
        workspace = self._api.workspaces.create(
            self._get_ws_without_vcs_create_payload())["data"]
        ws_id = workspace["id"]
        listed_ws = self._api.workspaces.list(page=0, page_size=50)["data"]
        found_ws = False
        for workspace in listed_ws:
            if workspace["id"] == ws_id:
                found_ws = True
                break
        self.assertTrue(found_ws)

        all_ws = self._api.workspaces.list_all()
        found_ws = False
        for workspace in all_ws:
            if workspace["id"] == ws_id:
                found_ws = True
                break
        self.assertTrue(found_ws)

        # Lock the workspace and confirm it's locked
        ws_locked = self._api.workspaces.lock(
            ws_id, {"reason": "Unit testing."})["data"]
        self.assertTrue(ws_locked["attributes"]["locked"])

        # Unlock the workspace and confirm it's unlocked
        ws_unlocked = self._api.workspaces.unlock(ws_id)
        self.assertFalse(ws_unlocked["data"]["attributes"]["locked"])

        # Relock the workspace and confirm it's locked
        ws_relocked = self._api.workspaces.lock(
            ws_id, {"reason": "Unit testing."})["data"]
        self.assertTrue(ws_relocked["attributes"]["locked"])

        # Force an unlock and confirm its unlocked
        ws_forced = self._api.workspaces.force_unlock(ws_id)["data"]
        self.assertFalse(ws_forced["attributes"]["locked"])

        # Update the workspace, check that the updates took effect
        updated_name = self._unittest_random_name()
        update_payload = {
            "data": {
                "type": "workspaces",
                "attributes": {
                    "name": updated_name
                }
            }
        }
        ws_updated = self._api.workspaces.update(update_payload, workspace_id=ws_id)["data"]
        self.assertEqual(
            updated_name, ws_updated["attributes"]["name"])

        # Assign an SSH key and confirm it's added
        assign_payload = {
            "data": {
                "attributes": {
                    "id": self._created_key_id
                },
                "type": "workspaces"
            }
        }
        self._api.workspaces.assign_ssh_key(ws_id, assign_payload)

        # Show by ID, and make sure we get the right payload back. Also check for the
        # the newly assigned SSH key
        ws_shown_by_id = self._api.workspaces.show(workspace_id=ws_id)["data"]
        self.assertEqual(ws_id, ws_shown_by_id["id"])
        self.assertIn("ssh-key", ws_shown_by_id["relationships"])

        # Unassign the SSH key and confirm it's removed
        unassign_payload = {
            "data": {
                "attributes": {
                    "id": None
                },
                "type": "workspaces"
            }
        }
        self._api.workspaces.unassign_ssh_key(ws_id, unassign_payload)
        ws_shown_by_id = self._api.workspaces.show(workspace_id=ws_id)["data"]
        self.assertNotIn("ssh-key", ws_shown_by_id["relationships"])

        self._api.workspaces.destroy(
            workspace_name=updated_name)
        listed_ws = self._api.workspaces.list()["data"]
        found_ws = False
        for workspace in listed_ws:
            if workspace["id"] == ws_id:
                found_ws = True
                break

        self.assertFalse(found_ws)



"""
Module for Terraform Cloud API Endpoint: Run Triggers.
"""

from .endpoint import TFCEndpoint
from ._constants import MAX_PAGE_SIZE

class TFCRunTriggers(TFCEndpoint):
    """
    `Run Triggers API Docs \
        <https://www.terraform.io/docs/cloud/api/run-triggers.html>`_
    """
    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/run-triggers"
        self._ws_api_v2_base_url = f"{self._api_v2_base_url}/workspaces"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, workspace_id, payload):
        """
        ``POST /workspaces/:workspace_id/run-triggers``

        `Run Triggers Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-triggers.html#create-a-run-trigger>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/run-triggers.html#sample-payload>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/run-triggers"
        return self._create(url, payload)

    def list(self, workspace_id, filters=None, page=None, page_size=None):
        """
        ``GET /workspaces/:workspace_id/run-triggers``

        `Run Triggers List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-triggers.html#list-run-triggers>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/run-triggers.html#query-parameters>`_):
            - ``filter[run-trigger][type]`` (Required)
            - ``page_size`` (Optional)

        Example filter(s):

        .. code-block:: python

            filters = [
                {
                    "keys": ["run-trigger", "type"],
                    "value": "foo"
                }
            ]
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/run-triggers"
        return self._list(url, filters=filters, page=page, page_size=page_size)

    def list_all(self, workspace_id, filters=None):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every run trigger for a workspace.

        Returns an array of objects.
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/run-triggers"

        current_page_number = 1
        run_triggers_resp = \
            self._list(url, page=current_page_number, page_size=MAX_PAGE_SIZE, filters=filters)
        total_pages = run_triggers_resp["meta"]["pagination"]["total-pages"]

        run_triggers = []
        while current_page_number <= total_pages:
            run_triggers_resp = \
                self._list(url, page=current_page_number, page_size=MAX_PAGE_SIZE, filters=filters)
            run_triggers += run_triggers_resp["data"]
            current_page_number += 1

        return run_triggers

    def show(self, run_trigger_id):
        """
        ``GET /run-triggers/:run_trigger_id``

        `Run Triggers Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-triggers.html#show-a-run-trigger>`_
        """
        url = f"{self._endpoint_base_url}/{run_trigger_id}"
        return self._show(url)

    def destroy(self, run_trigger_id):
        """
        ``DELETE /run-triggers/:run_trigger_id``

        `Run Triggers Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/run-triggers.html#delete-a-run-trigger>`_
        """
        url = f"{self._endpoint_base_url}/{run_trigger_id}"
        return self._destroy(url)



"""
Constants for default values on the TFC API endpoints
"""

from enum import Enum
import logging

# Default Config Items
TFC_SAAS_HOSTNAME = "app.terraform.io"
TFC_SAAS_URL = "https://app.terraform.io"
API_LOG_LEVEL = logging.CRITICAL
MAX_PAGE_SIZE = 100

# Common TFC API HTTP Codes
HTTP_OK = 200
HTTP_CREATED = 201
HTTP_ACCEPTED = 202
HTTP_NO_CONTENT = 204
HTTP_MOVED_TEMPORARILY = 302
HTTP_NOT_MODIFIED = 304
HTTP_TEMPORARY_REDIRECT = 307
HTTP_NOT_MODIFIED = 304
HTTP_BAD_REQUEST = 400
HTTP_UNAUTHORIZED = 401
HTTP_FORBIDDEN = 403
HTTP_NOT_FOUND = 404
HTTP_CONFLICT = 409
HTTP_PRECONDITION_FAILED = 412
HTTP_UNPROCESSABLE_ENTITY = 422
HTTP_API_REQUEST_RATE_LIMIT_REACHED = 429
HTTP_INTERNAL_SERVER_ERROR = 500


class Entitlements(Enum):
    """
    Enumeration of all the possible Terraform Cloud Entitlements, to be used
    on each endpoint to indicate the required entitlements to use that piece
    of the Terraform Cloud API.
    """

    AGENTS = "AGENTS"
    AUDIT_LOGGING = "AUDIT_LOGGING"
    CONFIGURATION_DESIGNER = "CONFIGURATION_DESIGNER"
    COST_ESTIMATION = "COST_ESTIMATION"
    OPERATIONS = "OPERATIONS"
    PRIVATE_MODULE_REGISTRY = "PRIVATE_MODULE_REGISTRY"
    PRIVATE_NETWORKING = "PRIVATE_NETWORKING"
    SELF_SERVE_BILLING = "SELF_SERVE_BILLING"
    SENTINEL = "SENTINEL"
    STATE_STORAGE = "STATE_STORAGE"
    SSO = "SSO"
    TEAMS = "TEAMS"
    USER_LIMIT = "USER_LIMIT"
    USAGE_REPORTING = "USAGE_REPORTING"
    VCS_INTEGRATIONS = "VCS_INTEGRATIONS"



"""
Module for Terraform Cloud API Endpoint: Orgs.
"""

from .endpoint import TFCEndpoint

class TFCOrgs(TFCEndpoint):
    """
    `Orgs API Docs \
        <https://www.terraform.io/docs/cloud/api/organizations.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST /organizations``

        `Orgs Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organizations.html#create-an-organization>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/organizations.html#sample-payload>`_
        """
        return self._create(self._org_api_v2_base_url, payload)

    def destroy(self, org_name):
        """
        ``DELETE /organizations/:organization_name``

        `Orgs Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organizations.html#destroy-an-organization>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}"
        return self._destroy(url)

    def entitlements(self, org_name):
        """
        ``GET /organizations/:organization_name/entitlement-set``

        `Orgs Entitlements API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organizations.html#show-the-entitlement-set>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}/entitlement-set"
        return self._get(url)

    def subscription(self, org_name):
        """
        ``GET /organizations/:organization_name/subscription``

        This endpoint is not currently documented in the TFC API docs, but it can be used to pull
        detailed subscription information for a TFC organization.
        """
        url = f"{self._org_api_v2_base_url}/{org_name}/subscription"
        return self._get(url)

    def list(self):
        """
        ``GET /organizations``

        `Orgs List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organizations.html#list-organizations>`_
        """
        return self._list(self._org_api_v2_base_url)

    def show(self, org_name):
        """
        ``GET /organizations/:organization_name``

        `Orgs Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organizations.html#show-an-organization>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}"
        return self._show(url)

    def update(self, org_name, payload):
        """
        ``PATCH /organizations/:organization_name``

        `Orgs Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organizations.html#update-an-organization>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/organizations.html#sample-payload-1>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}"
        return self._update(url, payload)



"""
Module for testing the Terraform Cloud API Endpoint: Invoices.
"""

from .base import TestTFCBaseTestCase


class TestTFCInvoices(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Invoices.
    """

    _unittest_name = "invoi"
    _endpoint_being_tested = "invoices"

    def test_invoices(self):
        """
        Test the Invoices API endpoints.
        """
        # TODO: this will fail since I don't have any billing info associated with the test org.
        # invoices = self._api.invoices.list()
        # print(invoices)



"""
Module for testing the Terraform Cloud API Endpoint: Policy Sets.
"""

import time
from .base import TestTFCBaseTestCase


class TestTFCPolicySets(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Policy Sets.
    """

    _unittest_name = "pol-sets"
    _endpoint_being_tested = "policy_sets"

    def setUp(self):
        # Set up a workspace to attach a policy set to
        workspace = self._api.workspaces.create(self._get_ws_without_vcs_create_payload())["data"]
        self._ws_id = workspace["id"]
        self._ws_name = workspace["attributes"]["name"]

        # Set up a policy to add and remove from the set
        policy = self._api.policies.create(self._get_policy_create_payload())["data"]
        self._policy_id = policy["id"]

    def tearDown(self):
        # Destroy the workspace and policy we created
        self._api.workspaces.destroy(
            workspace_name=self._ws_name)
        self._api.policies.destroy(self._policy_id)

    def test_policy_sets(self):
        """
        Test the Policy Sets API endpoints.
        """
        # List all the policy sets, confirm there are none
        all_policy_sets = self._api.policy_sets.list()["data"]
        self.assertEqual(len(all_policy_sets), 0)

        # Add a policy set to TFC
        created_policy_set = self._api.policy_sets.create(self._get_policy_create_payload())["data"]
        created_policy_set_id = created_policy_set["id"]
        created_policy_set_name = created_policy_set["attributes"]["name"]

        # Check that we now have one policy set, use the query options to ensure
        # that they work.
        test_filters = [
            {
                "keys": ["versioned"],
                "value": "false"
            }
        ]
        some_policy_sets = self._api.policy_sets.list(\
            filters=test_filters, page=0, page_size=50, \
            include="policies", search=created_policy_set_name)["data"]
        self.assertEqual(len(some_policy_sets), 1)

        all_policy_sets = self._api.policy_sets.list_all(\
            filters=test_filters, include="policies", \
                search=created_policy_set_name)
        self.assertEqual(len(all_policy_sets), 1)

        # Update the policy set, confirm the update took place
        desc_to_update_to = "foo"
        update_payload = {
            "data": {
                "attributes": {
                    "description": desc_to_update_to
                },
                "type": "policy-sets"
            }
        }
        updated_policy_set = self._api.policy_sets.update( \
            created_policy_set_id, update_payload)["data"]
        self.assertEqual(desc_to_update_to, updated_policy_set["attributes"]["description"])

        # Add the policy we created in the set up to the policy set
        add_remove_policy_payload = {
            "data": [
                {
                    "id": self._policy_id, "type": "policies"
                },
            ]
        }
        self._api.policy_sets.add_policies_to_set(created_policy_set_id, add_remove_policy_payload)
        shown_policy_set = self._api.policy_sets.show(created_policy_set_id)["data"]
        shown_policies_in_set = shown_policy_set["relationships"]["policies"]["data"]
        self.assertEqual(len(shown_policies_in_set), 1)

        # Attach the policy set to the workspace we created, confirm it's attached
        attach_detach_to_workspace_payload = {
            "data": [
                {
                    "id": self._ws_id, "type": "workspaces"
                },
            ]
        }
        self._api.policy_sets.attach_policy_set_to_workspaces(\
            created_policy_set_id, attach_detach_to_workspace_payload)
        shown_policy_set = self._api.policy_sets.show(created_policy_set_id)["data"]
        shown_workspaces_attached_to = \
            shown_policy_set["relationships"]["workspaces"]["data"]
        self.assertEqual(len(shown_workspaces_attached_to), 1)

        # Detach the policy set from the workspace we created, confirm it's not attached
        self._api.policy_sets.detach_policy_set_from_workspaces(\
            created_policy_set_id, attach_detach_to_workspace_payload)
        shown_policy_set = self._api.policy_sets.show(created_policy_set_id)["data"]
        self.assertEqual(len(shown_policy_set["relationships"]["workspaces"]["data"]), 0)

        # Remove the policy from the set and confirm it has been removed
        self._api.policy_sets.remove_policies_from_set(\
            created_policy_set_id, add_remove_policy_payload)
        shown_policy_set = self._api.policy_sets.show(created_policy_set_id)["data"]
        shown_policies_in_set = shown_policy_set["relationships"]["policies"]["data"]
        self.assertEqual(len(shown_policies_in_set), 0)

        # Delete the policy set, confirm it's been deleted
        self._api.policy_sets.destroy(created_policy_set_id)
        all_policy_sets = self._api.policy_sets.list()["data"]
        self.assertEqual(len(all_policy_sets), 0)


    def test_policy_sets_versions(self):
        """
        Test the Policy Sets Versions API endpoints.
        """
        # Create a new policy set and policy set version, without being attached to VCS.
        created_policy_set = self._api.policy_sets.create(self._get_policy_create_payload())["data"]
        created_policy_set_id = created_policy_set["id"]
        pol_set_version = self._api.policy_sets.create_policy_set_version(\
            created_policy_set_id)["data"]
        pol_set_version_id = pol_set_version["id"]

        # Upload the policy set tarball from testdata
        self._api.policy_sets.upload(
            self._policy_set_upload_tarball_path, pol_set_version_id)
        shown_pol_set_version = self._api.policy_sets.show_policy_set_version(\
            pol_set_version["id"])["data"]

        # Confirm the upload is completed and the policy set is ready to use
        time.sleep(5)
        ready_or_pending = \
            shown_pol_set_version["attributes"]["status"] in ["pending", "ready"]
        self.assertTrue(ready_or_pending)
        self._api.policy_sets.destroy(created_policy_set_id)



"""
Module for testing the Terraform Cloud API Endpoint: Plans.
"""

import time
import os

from .base import TestTFCBaseTestCase


class TestTFCPlans(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Plans.
    """

    _unittest_name = "plans"
    _endpoint_being_tested = "plans"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        # Store the OAuth client ID to remove it at the end.
        oauth_client = self._api.oauth_clients.create(self._get_oauth_client_create_payload())
        self._oauth_client_id = oauth_client["data"]["id"]
        oauth_token_id = oauth_client["data"]["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a workspace using that token ID, save the workspace ID
        _ws_payload = self._get_ws_with_vcs_create_payload(oauth_token_id)
        workspace = self._api.workspaces.create(_ws_payload)["data"]
        self._ws_id = workspace["id"]

        # Configure the required variables on the workspace
        variable_payloads = [
            self._get_variable_create_payload(
                "email", self._test_email, self._ws_id),
            self._get_variable_create_payload(
                "org_name", self._test_org_name, self._ws_id),
            self._get_variable_create_payload(
                "TFE_TOKEN", self._test_api_token, self._ws_id, category="env", sensitive=True)
        ]
        for payload in variable_payloads:
            self._api.vars.create(payload)

        # Sleep for 3 seconds to give the WS time to create, takes a bit longer
        # on small TFE instances.
        time.sleep(3)

        # Start the run, store the run ID
        create_run_payload = self._get_run_create_payload(self._ws_id)
        self._run = self._api.runs.create(create_run_payload)["data"]
        self._run_id = self._run["id"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_plans(self):
        """
        Test the Plans API endpoints.
        """

        # Create a run and wait for the created run to complete it's plan
        created_run = self._api.runs.show(self._run_id)["data"]
        created_plan_id = created_run["relationships"]["plan"]["data"]["id"]

        # Timeout if the plan doesn't reach confirmable, this can happen
        # if the run is queued.
        created_run = self._created_run_timeout(self._run_id)

        # Confirm the shown plan's ID matches the ID from the run
        plan = self._api.plans.show(created_plan_id)["data"]
        self.assertEqual(created_plan_id, plan["id"])

        if os.path.exists(self._plan_json_tarball_target_path):
            os.remove(self._plan_json_tarball_target_path)

        # Download the plan JSON by run ID, then assert the file exists, and remove it
        self._api.plans.download_json(self._plan_json_tarball_target_path, run_id=self._run_id)
        self.assertTrue(os.path.exists(self._plan_json_tarball_target_path))
        os.remove(self._plan_json_tarball_target_path)
        self.assertFalse(os.path.exists(self._plan_json_tarball_target_path))

        # Show the plan, assert that the Plan IDs match
        shown_plan = self._api.plans.show(created_plan_id)["data"]
        self.assertEqual(created_plan_id, shown_plan["id"])

        # Download the plan JSON by run ID, then assert the file exists, and remove it
        self._api.plans.download_json(self._plan_json_tarball_target_path, plan_id=plan["id"])
        self.assertTrue(os.path.exists(self._plan_json_tarball_target_path))
        os.remove(self._plan_json_tarball_target_path)
        self.assertFalse(os.path.exists(self._plan_json_tarball_target_path))



"""
Module for testing the Terraform Cloud API Endpoint: Admin Users.
"""

from .base import TestTFCBaseTestCase


class TestTFCAdminUsers(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Admin Users.
    """

    _unittest_name = "users"
    _endpoint_being_tested = "admin_users"

    def test_admin_users(self):
        """
        Test the Admin Users API endpoints.

        NOTE: Not tested, as they can't be reverted via the API: ``destroy``,
        ``disable_two_factor``.

        NOTE: Not tested, as it's not super valuable via the API: ``impersonate``,
        ``unimpersonate``.
        """

        # List all the users through the admin users API, but query for our test
        # user
        listed_users = self._api.admin_users.list(\
            query=self._test_username, page=0, page_size=50)["data"]
        found_user = False
        for user in listed_users:
            if self._test_username == user["attributes"]["username"]:
                found_user = True
                break
        self.assertTrue(found_user)

        # Extract the test users's ID
        test_user_id = listed_users[0]["id"]

        # Suspend the user, confirm they are suspended
        suspended_user = self._api.admin_users.suspend(test_user_id)["data"]
        self.assertTrue(suspended_user["attributes"]["is-suspended"])

        # Unsuspend the user, confirm they are no longer suspended
        unsuspended_user = self._api.admin_users.unsuspend(test_user_id)["data"]
        self.assertFalse(unsuspended_user["attributes"]["is-suspended"])

        # Grant the user admin rights, confirm they are now admins
        admin_user = self._api.admin_users.grant_admin(test_user_id)["data"]
        self.assertTrue(admin_user["attributes"]["is-admin"])

        # Revoke the user's admin rights, confirm they are no longer admins
        nonadmin_user = self._api.admin_users.revoke_admin(test_user_id)["data"]
        self.assertFalse(nonadmin_user["attributes"]["is-admin"])



"""
Module for Terraform Cloud API Endpoint: Teams.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCTeams(TFCEndpoint):
    """
    `Teams API Docs \
        <https://www.terraform.io/docs/cloud/api/teams.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._teams_api_v2_base_url = f"{self._api_v2_base_url}/teams"
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations/{org_name}/teams"

    def required_entitlements(self):
        return [Entitlements.TEAMS]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST /organizations/:organization_name/teams``

        `Teams Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/teams.html#create-a-team>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/teams.html#sample-payload>`_
        """
        return self._create(self._org_api_v2_base_url, payload)

    def destroy(self, team_id):
        """
        ``DELETE /teams/:team_id``

        `Teams Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/teams.html#delete-a-team>`_
        """
        url = f"{self._teams_api_v2_base_url}/{team_id}"
        return self._destroy(url)

    def list(self):
        """
        ``GET organizations/:organization_name/teams``

        `Teams List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/teams.html#list-teams>`_
        """
        return self._list(self._org_api_v2_base_url)

    def show(self, team_id):
        """
        ``GET /teams/:team_id``

        `Teams Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/teams.html#show-team-information>`_
        """
        url = f"{self._teams_api_v2_base_url}/{team_id}"
        return self._show(url)

    def update(self, team_id, payload):
        """
        ``PATCH /teams/:team_id``

        `Teams Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/teams.html#update-a-team>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/teams.html#sample-payload-1>`_
        """
        url = f"{self._teams_api_v2_base_url}/{team_id}"
        return self._update(url, payload)



"""
Module for testing the Terraform Cloud API Endpoint: Orgs.
"""

from .base import TestTFCBaseTestCase


class TestTFCOrgs(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Orgs.
    """

    _unittest_name = "orgs"
    _endpoint_being_tested = "orgs"

    def test_orgs(self):
        """
        Test the Orgs API endpoints.
        """

        # Create an org, confirm it was created.
        created_org = self._api.orgs.create(self._get_org_create_payload())["data"]
        created_org_id = created_org["id"]
        created_org_name = created_org["attributes"]["name"]
        all_orgs = self._api.orgs.list()["data"]
        self.assertNotEqual(
            len(all_orgs), 0, msg="No organizations found for TFC token.")

        # List entitlements, confirm expected response
        ent = self._api.orgs.entitlements(self._test_org_name)["data"]
        self.assertEqual(ent["type"], "entitlement-sets")
        self.assertTrue(ent["attributes"]["state-storage"])

        # List subscription details, confirm expected response
        if self._api.is_terraform_cloud():
            # This only works on TFC, so skip it if we're on TFE.
            sub = self._api.orgs.subscription(self._test_org_name)["data"]
            if "included" in sub:
                # If "included" is present, it's an active subscription
                self.assertEqual(sub["type"], "subscriptions")
                self.assertIn("identifier", sub["included"][0]["attributes"])
            else:
                # Otherwise, make sure it's a free tier subscription
                self.assertTrue(sub["attributes"]["is-public-free-tier"])

        # Show the org, confirm IDs match
        org = self._api.orgs.show(self._test_org_name)
        self.assertEqual(org["data"]["id"], self._test_org_name)

        # Change the email address for the org, confirm the change.
        email_name = self._unittest_random_name()
        email_to_update_to = f"{email_name}@gmail.com"
        update_org_payload = {
            "data": {
                "type": "organizations",
                "attributes": {
                    "email": email_to_update_to
                }
            }
        }
        updated_org = self._api.orgs.update(
            self._test_org_name, update_org_payload)
        self.assertEqual(updated_org["data"] \
                         ["attributes"]["email"], email_to_update_to)

        # Destroy the org, confirm it's gone
        self._api.orgs.destroy(created_org_name)
        all_orgs = self._api.orgs.list()["data"]
        found_org = False
        for org in all_orgs:
            if org["id"] == created_org_id:
                found_org = True
                break
        self.assertFalse(found_org)



"""
Module for Terraform Cloud API Endpoint: Config Versions.
"""
import io
import tarfile
from .endpoint import TFCEndpoint
from ._constants import MAX_PAGE_SIZE

class TFCConfigVersions(TFCEndpoint):
    """
    `Config Versions API Docs \
        <https://www.terraform.io/docs/cloud/api/configuration-versions.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._ws_api_v2_base_url = f"{self._api_v2_base_url}/workspaces"
        self._config_version_api_v2_base_url = f"{self._api_v2_base_url}/configuration-versions"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def list(self, workspace_id, page=None, page_size=None):
        """
        ``GET /workspaces/:workspace_id/configuration-versions``

        `Config Versions List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/configuration-versions.html#list-configuration-versions>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/configuration-versions.html#query-parameters>`_):
            - ``page`` (Optional)
            - ``page_size`` (Optional)
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/configuration-versions"
        return self._list(url, page=page, page_size=page_size)

    def list_all(self, workspace_id):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every config version in a workspace.

        Returns an array of objects.
        """
        current_page_number = 1
        config_versions_resp = \
            self.list(workspace_id, page=current_page_number, page_size=MAX_PAGE_SIZE)
        total_pages = config_versions_resp["meta"]["pagination"]["total-pages"]

        config_versions = []
        while current_page_number <= total_pages:
            config_versions_resp = \
                self.list(workspace_id, page=current_page_number, page_size=MAX_PAGE_SIZE)
            config_versions += config_versions_resp["data"]
            current_page_number += 1

        return config_versions

    def show(self, config_version_id):
        """
        ``GET /configuration-versions/:configuration-id``

        `Config Versions Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/configuration-versions.html#show-a-configuration-version>`_
        """
        url = f"{self._config_version_api_v2_base_url}/{config_version_id}"
        return self._show(url)

    def create(self, workspace_id, payload):
        """
        ``POST /workspaces/:workspace_id/configuration-versions``

        `Config Versions Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/configuration-versions.html#create-a-configuration-version>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/configuration-versions.html#sample-payload>`_
        """
        url = f"{self._ws_api_v2_base_url}/{workspace_id}/configuration-versions"
        return self._create(url, payload)

    def upload(self, path_to_tarball, upload_url):
        """
        ``PUT https://archivist.terraform.io/v1/object/<UNIQUE OBJECT ID>``

        `Config Versions Upload API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/configuration-versions.html#upload-configuration-files>`_
        """
        data = None

        with open(path_to_tarball, 'rb') as data_bytes:
            data = data_bytes.read()

        return self._put(upload_url, data=data)

    def upload_from_string(self, template_string, upload_url):
        """
        ``PUT https://archivist.terraform.io/v1/object/<UNIQUE OBJECT ID>``

        `Config Versions Upload API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/configuration-versions.html#upload-configuration-files>`_

        Set configuration version from string, rather than pre-existing tarball.

        NOTE: this does not map to typical API usage, but for ease of use in some use cases,
        it's fine.
        """

        # create template io obj
        template_data = template_string.encode('utf-8')
        template_io = io.BytesIO(template_data)

        # create tarfile io obj
        targz_io = io.BytesIO()

        # add data to targz
        with tarfile.open(fileobj=targz_io, mode='w:gz') as tar:

            # create tarinfo object w/ desired path and size
            tarinfo = tarfile.TarInfo("main.tf")
            tarinfo.size = len(template_data)

            # return cursor to 0
            template_io.seek(0)

            # add the prepared item
            tar.addfile(tarinfo, template_io)

        targz_io.seek(0)

        # upload the template
        return self._put(upload_url, data=targz_io.read())



"""
Module for Terraform Cloud API Endpoint: Admin Users.
"""

from .endpoint import TFCEndpoint

class TFCAdminUsers(TFCEndpoint):
    """
    `Admin Users API Docs \
        <https://www.terraform.io/docs/cloud/api/admin/users.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/admin/users"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return True

    def destroy(self, user_id):
        """
        ``DELETE /admin/users/:id``

        `Admin Users Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/users.html#delete-a-user-account>`_
        """
        url = f"{self._endpoint_base_url}/{user_id}"
        return self._destroy(url)

    def disable_two_factor(self, user_id):
        """
        ``POST /admin/users/:id/actions/disable_two_factor``

        `Admin Users Disable Two Factor API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/users.html#disable-a-user-39-s-two-factor-authentication>`_
        """
        url = f"{self._endpoint_base_url}/{user_id}/actions/disable_two_factor"
        return self._post(url)

    def grant_admin(self, user_id):
        """
        ``POST /admin/users/:id/actions/grant_admin``

        `Admin Users Grant Admin API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/users.html#grant-a-user-administrative-privileges>`_
        """
        url = f"{self._endpoint_base_url}/{user_id}/actions/grant_admin"
        return self._post(url)

    def impersonate(self, user_id):
        """
        ``POST /admin/users/:id/actions/impersonate``

        `Admin Users Impersonate API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/users.html#impersonate-another-user>`_
        """
        url = f"{self._endpoint_base_url}/{user_id}/actions/impersonate"
        return self._post(url)

    def list(self, query=None, filters=None, page=None, page_size=None):
        """
        ``GET /admin/users``

        `Admin Users List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/users.html#list-all-users>`_
        """
        return self._list(\
            self._endpoint_base_url, query=query, filters=filters, page=page, page_size=page_size)

    def revoke_admin(self, user_id):
        """
        ``POST /admin/users/:id/actions/revoke_admin``

        `Admin Users Revoke Admin API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/users.html#revoke-an-user-39-s-administrative-privileges>`_
        """
        url = f"{self._endpoint_base_url}/{user_id}/actions/revoke_admin"
        return self._post(url)

    def suspend(self, user_id):
        """
        ``POST /admin/users/:id/actions/suspend``

        `Admin Users Suspend API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/users.html#suspend-a-user>`_
        """
        url = f"{self._endpoint_base_url}/{user_id}/actions/suspend"
        return self._post(url)

    def unimpersonate(self):
        """
        ``POST /admin/users/actions/unimpersonate``

        `Admin Users Unimpersonate API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/users.html#end-an-impersonation-session>`_
        """
        url = f"{self._endpoint_base_url}/actions/unimpersonate"
        return self._post(url)

    def unsuspend(self, user_id):
        """
        ``POST /admin/users/:id/actions/unsuspend``

        `Admin Users Unsuspend API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/users.html#re-activate-a-suspended-user>`_
        """
        url = f"{self._endpoint_base_url}/{user_id}/actions/unsuspend"
        return self._post(url)



"""
Module for testing the Terraform Cloud API Endpoint: Account.
"""

from .base import TestTFCBaseTestCase


class TestTFCAccount(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Account.
    """

    _unittest_name = "acc"
    _endpoint_being_tested = "account"

    def test_account(self):
        """
        Test the Account API endpoints.
        """

        # Show the account and confirm we get a valid response
        shown_account = self._api.account.show()["data"]
        self.assertIn("id", shown_account)

        # Store the original username and email for later usage
        original_email = shown_account["attributes"]["email"]
        original_username = shown_account["attributes"]["username"]

        # Generate a new email to change on the account, confirm the update
        email_name = self._unittest_random_name()
        email_to_update_to = f"{email_name}@gmail.com"
        update_payload = {
            "data": {
                "type": "users",
                "attributes": {
                    "email": email_to_update_to,
                    "username": original_username
                }
            }
        }
        updated_account = self._api.account.update(update_payload)["data"]

        # If it's TFC, we don't want to actually change the email, so we will
        # check against the unconfirmed email.
        if not self._api.is_terraform_cloud():
            updated_email = updated_account["attributes"]["email"]
        else:
            updated_email = updated_account["attributes"]["unconfirmed-email"]

        self.assertEqual(updated_email, email_to_update_to)

        # Set it back to the original email
        update_payload = {
            "data": {
                "type": "users",
                "attributes": {
                    "email": original_email,
                    "username": original_username
                }
            }
        }
        updated_account = self._api.account.update(update_payload)["data"]
        updated_email = updated_account["attributes"]["email"]
        self.assertEqual(updated_email, original_email)

        # Only fiddle around with the password for the account if it's not TFC,
        # and is a TFE instance used only for testing.
        if not self._api.is_terraform_cloud():
            # Update the password and confirm the request didn't fail
            password_to_update_to = self._unittest_random_name()
            change_password_payload = {
                "data": {
                    "type": "users",
                    "attributes": {
                        "current_password": self._test_password,
                        "password": password_to_update_to,
                        "password_confirmation": password_to_update_to
                    }
                }
            }
            changed_account = self._api.account.change_password(change_password_payload)["data"]
            self.assertIn("id", changed_account)

            # Change the password back to the initial one to make this easier
            change_password_payload = {
                "data": {
                    "type": "users",
                    "attributes": {
                        "current_password": password_to_update_to,
                        "password": self._test_password,
                        "password_confirmation": self._test_password
                    }
                }
            }
            changed_account = self._api.account.change_password(change_password_payload)["data"]
            self.assertIn("id", changed_account)



"""
Module for Terraform Cloud API Endpoint: Agent Tokens.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCAgentTokens(TFCEndpoint):
    """
    `Agent Tokens API Docs \
        <https://www.terraform.io/docs/cloud/api/agent-tokens.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._agent_pools_api_v2_base_url = f"{self._api_v2_base_url}/agent-pools"
        self._auth_tokens_api_v2_base_url = f"{self._api_v2_base_url}/authentication-tokens"

    def required_entitlements(self):
        return [Entitlements.AGENTS]

    def terraform_cloud_only(self):
        return True

    def terraform_enterprise_only(self):
        return True

    def create(self, agent_pool_id, payload):
        """
        ``POST /agent-pools/:agent_pool_id/authentication-tokens``

        `Agent Tokens Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agent-tokens.html#create-an-agent-token>`_
        """
        url = f"{self._agent_pools_api_v2_base_url}/{agent_pool_id}/authentication-tokens"
        return self._create(url, payload)

    def list(self, agent_pool_id):
        """
        ``GET /agent-pools/:agent_pool_id/authentication-tokens``

        `Agent Tokens List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agent-tokens.html#list-agent-tokens>`_
        """
        url = f"{self._agent_pools_api_v2_base_url}/{agent_pool_id}/authentication-tokens"
        return self._list(url)

    def show(self, token_id):
        """
        ``GET /authentication-tokens/:id``

        `Agent Tokens Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agent-tokens.html#show-an-agent-token>`_
        """
        url = f"{self._auth_tokens_api_v2_base_url}/{token_id}"
        return self._show(url)

    def destroy(self, token_id):
        """
        ``DELETE /api/v2/authentication-tokens/:id``

        `Agent Tokens Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agent-tokens.html#destroy-an-agent-token>`_
        """
        url = f"{self._auth_tokens_api_v2_base_url}/{token_id}"
        return self._destroy(url)



"""
Module for Terraform Cloud API Endpoint: Admin Terraform Versions.
"""

from .endpoint import TFCEndpoint

class TFCAdminTerraformVersions(TFCEndpoint):
    """
    `Admin Terraform Versions API Docs \
        <https://www.terraform.io/docs/cloud/api/admin/terraform-versions.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/admin/terraform-versions"

    def list(self, page=None, page_size=None):
        """
        ``GET /admin/terraform-versions``

        `Admin Terraform Versions List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/terraform-versions.html#list-all-terraform-versions>`_
        """
        return self._list(self._endpoint_base_url, page=page, page_size=page_size)

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return True

    def create(self, data):
        """
        ``POST /admin/terraform-versions``

        `Admin Terraform Versions Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/terraform-versions.html#create-a-terraform-version>`_
        """
        return self._post(self._endpoint_base_url, data=data)

    def show(self, version_id):
        """
        ``GET /admin/terraform-versions/:id``

        `Admin Terraform Versions Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/terraform-versions.html#show-a-terraform-version>`_
        """
        url = f"{self._endpoint_base_url}/{version_id}"
        return self._show(url)

    def update(self, version_id, data):
        """
        ``PATCH /admin/terraform-versions/:id``

        `Admin Terraform Versions Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/terraform-versions.html#update-a-terraform-version>`_
        """
        url = f"{self._endpoint_base_url}/{version_id}"
        return self._patch(url, data)

    def destroy(self, version_id):
        """
        ``DELETE /admin/terraform-versions/:id``

        `Admin Terraform Versions Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/terraform-versions.html#delete-a-terraform-version>`_
        """
        url = f"{self._endpoint_base_url}/{version_id}"
        return self._destroy(url)



"""
Module for testing the Terraform Cloud API Endpoint: Runs.
"""

import time

from terrasnek.exceptions import TFCHTTPConflict
from .base import TestTFCBaseTestCase


class TestTFCRuns(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Runs.
    """

    _unittest_name = "runs"
    _endpoint_being_tested = "runs"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        # Store the OAuth client ID to remove it at the end.
        oauth_client = self._api.oauth_clients.create(self._get_oauth_client_create_payload())
        self._oauth_client_id = oauth_client["data"]["id"]
        oauth_token_id = oauth_client["data"]["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a workspace using that token ID, save the workspace ID
        _ws_payload = self._get_ws_with_vcs_create_payload(oauth_token_id)
        workspace = self._api.workspaces.create(_ws_payload)["data"]
        self._ws_id = workspace["id"]

        # Configure the required variables on the workspace
        variable_payloads = [
            self._get_variable_create_payload(
                "email", self._test_email, self._ws_id),
            self._get_variable_create_payload(
                "org_name", self._test_org_name, self._ws_id),
            self._get_variable_create_payload(
                "TFE_TOKEN", self._test_api_token, self._ws_id, category="env", sensitive=True)
        ]
        for payload in variable_payloads:
            self._api.vars.create(payload)

        # Sleep for 3 seconds to give the WS time to create, takes a bit longer
        # on small TFE instances.
        time.sleep(3)

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_run_and_apply(self):
        """
        Test the Runs API endpoints with an apply.
        """
        # Create a run
        create_payload = self._get_run_create_payload(self._ws_id)
        run = self._api.runs.create(create_payload)["data"]
        run_id = run["id"]

        # Timeout if the plan doesn't reach confirmable, this can happen
        # if the run is queued.
        created_run = self._created_run_timeout(run_id)

        self.assertEqual(created_run["relationships"]["workspace"]["data"]["id"],
                         create_payload["data"]["relationships"]["workspace"]["data"]["id"])
        self.assertTrue(created_run["attributes"]["actions"]["is-confirmable"], True)
        self.assertRaises(
            KeyError, lambda: created_run["attributes"]["status-timestamps"]["applying-at"])

        # List the runs, using the correct parameters, confirm it has been created
        some_runs = self._api.runs.list(self._ws_id, page=0, page_size=50)["data"]
        found_run = False
        for run in some_runs:
            if run["id"] == run_id:
                found_run = True
                break
        self.assertTrue(found_run)

        all_runs = self._api.runs.list_all(self._ws_id)
        found_run = False
        for run in all_runs:
            if run["id"] == run_id:
                found_run = True
                break
        self.assertTrue(found_run)

        # Apply the plan
        apply_payload = {
            "comment": "foo"
        }
        self._api.runs.apply(run_id, apply_payload)

        # Wait for the plan to apply, then confirm the apply took place
        status_timestamps = self._api.runs.show(run_id)["data"]["attributes"]["status-timestamps"]
        while "applying-at" not in status_timestamps:
            time.sleep(1)
            status_timestamps = \
                self._api.runs.show(run_id)["data"]["attributes"]["status-timestamps"]
        self.assertIsNotNone(status_timestamps["applying-at"])

    def test_run_and_discard(self):
        """
        Test the Runs API endpoints for a discard.
        """

        # Create a run
        create_payload = self._get_run_create_payload(self._ws_id)
        run = self._api.runs.create(create_payload)["data"]
        run_id = run["id"]

        created_run = self._api.runs.show(run_id)["data"]
        self.assertRaises(
            KeyError, lambda: created_run["attributes"]["status-timestamps"]["discarded-at"])

        # Timeout if the plan doesn't reach confirmable, this can happen
        # if the run is queued.
        created_run = self._created_run_timeout(run_id)

        # Discard the run
        discard_payload = {
            "comment": "foo"
        }
        self._api.runs.discard(run_id, discard_payload)
        status_timestamps = self._api.runs.show(run_id)["data"]["attributes"]["status-timestamps"]
        while "discarded-at" not in status_timestamps:
            time.sleep(1)
            status_timestamps = \
                self._api.runs.show(run_id)["data"]["attributes"]["status-timestamps"]
        self.assertIsNotNone(status_timestamps["discarded-at"])

    def test_run_and_cancel(self):
        """
        Test the Runs API endpoints for a normal cancel.
        """

        # Create a run
        create_payload = self._get_run_create_payload(self._ws_id)
        run = self._api.runs.create(create_payload)["data"]
        run_id = run["id"]

        # Show the created run, make sure it hasn't yet been cancelled
        created_run = self._api.runs.show(run_id)["data"]
        self.assertIsNone(created_run["attributes"]["canceled-at"])

        # Wait for it to plan
        self._logger.debug("Sleeping while plan half-executes...")
        time.sleep(1)
        self._logger.debug("Done sleeping.")

        # Cancel the run, confirm it has been cancelled
        cancel_payload = {
            "comment": "foo"
        }
        self._api.runs.cancel(run_id, cancel_payload)
        status_timestamps = self._api.runs.show(run_id)["data"]["attributes"]["status-timestamps"]
        while "force-canceled-at" not in status_timestamps:
            time.sleep(1)
            status_timestamps = \
                self._api.runs.show(run_id)["data"]["attributes"]["status-timestamps"]
        self.assertIsNotNone(status_timestamps["force-canceled-at"])

    def test_run_and_force_cancel(self):
        """
        Test the Runs API endpoints for a force cancel.
        """

        # Create a run
        create_payload = self._get_run_create_payload(self._ws_id)
        run = self._api.runs.create(create_payload)["data"]
        run_id = run["id"]

        # Show the created run, make sure it hasn't yet been cancelled
        created_run = self._api.runs.show(run_id)["data"]
        self.assertIsNone(created_run["attributes"]["canceled-at"])

        # Wait for it to plan
        self._logger.debug("Sleeping while plan half-executes...")
        time.sleep(1)
        self._logger.debug("Done sleeping.")

        # Cancel the run first, it has to be cancelled before it can be force cancelled
        cancel_payload = {
            "comment": "foo"
        }
        self._api.runs.cancel(run_id, cancel_payload)

        # Cancel the run, confirm it has been cancelled
        force_cancel_payload = {
            "comment": "foo"
        }
        self.assertRaises(\
            TFCHTTPConflict, self._api.runs.force_cancel, run_id, force_cancel_payload)

        run_attrs = self._api.runs.show(run_id)["data"]["attributes"]
        while "force-cancel-available-at" not in run_attrs:
            shown_run = self._api.runs.show(run_id)["data"]
            run_attrs = shown_run["attributes"]
            time.sleep(1)
        self.assertIsNotNone(run_attrs["force-cancel-available-at"])



"""
Module for Terraform Cloud API Endpoint: Team Memberships.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCTeamMemberships(TFCEndpoint):
    """
    `Team Membership API Docs \
        <https://www.terraform.io/docs/cloud/api/team-members.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/teams"

    def required_entitlements(self):
        return [Entitlements.TEAMS]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def add_user_to_team(self, team_id, payload):
        """
        ``POST /teams/:team_id/relationships/users``

        `Team Membership Add User To Team API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/team-members.html#add-a-user-to-team>`_

        `Add Sample Payload \
            <https://www.terraform.io/docs/cloud/api/team-members.html#sample-payload>`_
        """
        url = f"{self._endpoint_base_url}/{team_id}/relationships/users"
        return self._post(url, data=payload)

    def remove_user_from_team(self, team_id, payload):
        """
        ``DELETE /teams/:team_id/relationships/users``

        `Team Membership Remove User From Team API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/team-members.html#delete-a-user-from-team>`_

        `Remove Sample Payload \
            <https://www.terraform.io/docs/cloud/api/team-members.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{team_id}/relationships/users"
        return self._delete(url, data=payload)



"""
Module for testing the Terraform Cloud API Endpoint: Org Memberships.
"""

from .base import TestTFCBaseTestCase


class TestTFCOrgMemberships(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Org Memberships.
    """

    _unittest_name = "org-mem"
    _endpoint_being_tested = "org_memberships"

    def test_org_memberships(self):
        """
        Test the Org Memberships API endpoints.
        """

        # NOTE: User needs to be created ahead of time, and it can't be done with
        # the API currently.

        # Get the existing org memberships for the logged in user
        orgs_for_user = self._api.org_memberships.list_for_user()["data"]
        in_test_org = False
        for org in orgs_for_user:
            if self._test_org_name == org["relationships"]["organization"]["data"]["id"]:
                in_test_org = True
        self.assertTrue(in_test_org)

        # Invite a user
        invite_payload = self._get_org_membership_invite_payload()
        invite = self._api.org_memberships.invite(invite_payload)["data"]
        invited_email = invite["attributes"]["email"]
        self.assertEqual(invited_email, self._test_email)

        # Show a user'sc org membership using the org membership ID
        org_membership_id = invite["id"]
        shown_membership = self._api.org_memberships.show(org_membership_id)
        shown_membership_id = shown_membership["data"]["id"]
        self.assertEqual(org_membership_id, shown_membership_id)

        # List the users in the org, but do some filtering to test out those features,
        # and only look for the user we just added
        test_filters = [
            {
                "keys": ["status"],
                "value": "inactive"
            }
        ]
        truncated_test_username = self._test_username[:5]
        users_for_org = self._api.org_memberships.list_for_org(\
            query=truncated_test_username, filters=test_filters, page=0, page_size=50)
        num_users_in_org = len(users_for_org["data"])
        # Only one cause we are filtering on the test username (inactive users)
        self.assertEqual(num_users_in_org, 1)

        all_users_for_org = self._api.org_memberships.list_all_for_org(\
            query=truncated_test_username, filters=test_filters)
        num_users_in_org = len(all_users_for_org)
        # Only one cause we are filtering on the test username (inactive users)
        self.assertEqual(num_users_in_org, 1)

        # Remove the user, confirm that the user is gone when we filter on the username
        self._api.org_memberships.remove(org_membership_id)
        users_for_org = self._api.org_memberships.list_for_org(\
            query=truncated_test_username, filters=test_filters, page=0, page_size=50)
        num_users_in_org = len(users_for_org["data"])
        self.assertEqual(num_users_in_org, 0)



"""
Module containing all of the exception classes used for this API client.
"""

# https://docs.python.org/3/tutorial/errors.html

# Common Exceptions
class TFCException(Exception):
    """Base class for Terraform Cloud errors."""

# Terraform Cloud Specific Exceptions
class InvalidTFCTokenException(TFCException):
    """Cannot instantiate TFC API class without a valid TFC_TOKEN."""

class TFCDeprecatedWontFix(TFCException):
    """Terraform Cloud deprecated endpoint. Won't be fixed, use another endpoint."""

# HTTP Exceptions
class TFCHTTPBadRequest(TFCException):
    """Terraform Cloud bad request. (HTTP 400)

    The HTTP 400 Bad Request response status code indicates that the server
    cannot or will not process the request due to something that is perceived
    to be a client error (e.g., malformed request syntax, invalid request
    message framing, or deceptive request routing).
    """

class TFCHTTPUnauthorized(TFCException):
    """Terraform Cloud unauthorized. (HTTP 401)

    The HTTP 401 Unauthorized client error status response code indicates that
    the request has not been applied because it lacks valid authentication
    credentials for the target resource.
    """

class TFCHTTPForbidden(TFCException):
    """Terraform Cloud forbidden. (HTTP 403)

    The HTTP 403 Forbidden client error status response code indicates that
    the server understood the request but refuses to authorize it.
    """

class TFCHTTPNotFound(TFCException):
    """Terraform Cloud not found. (HTTP 404)

    The HTTP 404 Not Found client error response code indicates that the server
    can't find the requested resource. Links that lead to a 404 page are often
    called broken or dead links and can be subject to link rot.
    """

class TFCHTTPConflict(TFCException):
    """Terraform Cloud conflict. (HTTP 409)

    The HTTP 409 Conflict response status code indicates a request conflict with
    current state of the target resource.
    """

class TFCHTTPPreconditionFailed(TFCException):
    """Terraform Cloud precondition failed. (HTTP 412)

    The HTTP 412 Precondition Failed client error response code indicates that
    access to the target resource has been denied. This happens with
    conditional requests on methods other than GET or HEAD when the condition
    defined by the If-Unmodified-Since or If-None-Match headers is not fulfilled.
    In that case, the request, usually an upload or a modification of a resource,
    cannot be made and this error response is sent back.
    """

class TFCHTTPUnprocessableEntity(TFCException):
    """Terraform Cloud unprocessable entity. (HTTP 422)

    The HTTP 422 Unprocessable Entity response status code indicates that the
    server understands the content type of the request entity, and the syntax
    of the request entity is correct, but it was unable to process the contained
    instructions.
    """

class TFCHTTPAPIRequestRateLimit(TFCException):
    """Terraform Cloud API rate limit has been reached. (HTTP 429)

    The HTTP 429 Rate limit has been reached error indicates that the client has
    reached the set request limit for the organization. By default; terraform enterprise
    and cloud APIs have a default limit of 30 requests per second as as authenticated or
    unauthenticated request.
    """

class TFCHTTPInternalServerError(TFCException):
    """Terraform Cloud internal server error. (HTTP 500)

    The HTTP 500 Internal Server Error server error response code indicates that
    the server encountered an unexpected condition that prevented it from
    fulfilling the request.
    """

class TFCHTTPUnclassified(TFCException):
    """Terraform Cloud unclassified error.

    This is a catch-all class to manage any HTTP exceptions that are not
    enumerated.
    """



"""
Module for Terraform Cloud API Endpoint: Admin Orgs.
"""

from .endpoint import TFCEndpoint

class TFCAdminOrgs(TFCEndpoint):
    """
    `Admin Orgs API Docs \
        <https://www.terraform.io/docs/cloud/api/admin/organizations.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/admin/organizations"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return True

    def destroy(self, org_name):
        """
        ``DELETE /admin/organizations/:name``

        `Admin Destroy Org API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#delete-an-organization>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}"
        return self._destroy(url)

    def list(self):
        """
        ``GET /admin/organizations``

        `Admin List Orgs API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#list-all-organizations>`_

        This endpoint lists all organizations in the Terraform Cloud installation.
        """
        return self._list(self._org_api_v2_base_url)

    def show(self, org_name):
        """
        ``GET /admin/organizations/:name``

        `Admin Show Org API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#show-an-organization>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}"
        return self._show(url)

    def update(self, org_name, payload):
        """
        ``PATCH /admin/organizations/:organization_name``

        `Admin Orgs Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#update-an-organization>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/admin/organizations.html#sample-payload>`_
        """
        url = f"{self._org_api_v2_base_url}/{org_name}"
        return self._update(url, payload)



"""
Module for testing the Terraform Cloud API Endpoint: Team Memberships.
"""

from .base import TestTFCBaseTestCase


class TestTFCTeamMemberships(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Team Memberships.
    """

    _unittest_name = "team-mem"
    _endpoint_being_tested = "team_memberships"

    def setUp(self):
        self._team = self._api.teams.create(
            self._get_team_create_payload())["data"]
        self._team_id = self._team["id"]
        invite = self._api.org_memberships.invite(self._get_org_membership_invite_payload())
        self._org_membership_id = invite["data"]["id"]

    def tearDown(self):
        self._api.teams.destroy(self._team_id)
        self._api.org_memberships.remove(self._org_membership_id)

    def test_team_memberships(self):
        """
        Test the Team Memberships API endpoints.
        """

        logged_in_user = self._api.account.show()["data"]
        logged_in_username = logged_in_user["attributes"]["username"]

        # Add our user to the team, confirm they have been added
        membership_payload = {
            "data": [
                {
                    "type": "users",
                    "id": logged_in_username
                }
            ]
        }
        self._api.team_memberships.add_user_to_team(
            self._team_id, membership_payload)

        # Show the team memberships, confirm that only one user is in the teeam
        # since we just created it.
        shown_team = self._api.teams.show(self._team_id)["data"]
        self.assertEqual(len(shown_team["relationships"]["users"]["data"]), 1)

        # Remove the user from the team, and confirm there are no users
        # left in the team.
        self._api.team_memberships.remove_user_from_team(
            self._team_id, membership_payload)
        shown_team = self._api.teams.show(self._team_id)["data"]
        self.assertEqual(len(shown_team["relationships"]["users"]["data"]), 0)



"""
Module for testing the Terraform Cloud API Endpoint: Admin Orgs.
"""

from .base import TestTFCBaseTestCase


class TestTFCAdminOrgs(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Admin Orgs.
    """

    _unittest_name = "adm-org"
    _endpoint_being_tested = "admin_orgs"

    def setUp(self):
        # Create a temp org to manipulate in the test
        created_org = self._api.orgs.create(self._get_org_create_payload())["data"]
        self._created_org_name = created_org["attributes"]["name"]
        self._created_org_id = created_org["id"]

    def test_admin_orgs(self):
        """
        Test the Admin Orgs API endpoints.
        """

        # List all the orgs, confirm the created one is present
        all_orgs = self._api.admin_orgs.list()["data"]
        found_created_org = False
        for org in all_orgs:
            org_id = org["id"]
            if org_id == self._created_org_id:
                found_created_org = True
                break
        self.assertTrue(found_created_org)

        # Change the email address for the org, confirm the change.
        update_org_payload = {
            "data": {
                "type": "organizations",
                "attributes": {
                    "global-module-sharing": True
                }
            }
        }
        updated_org = self._api.admin_orgs.update(self._test_org_name, update_org_payload)["data"]
        self.assertTrue(updated_org["attributes"]["global-module-sharing"])

        # Show the created org, confirm it matches the created org's ID
        shown_org = self._api.admin_orgs.show(self._created_org_name)["data"]
        self.assertEqual(self._created_org_id, shown_org["id"])

        # Destroy the org that we created, verify it's gone.
        self._api.admin_orgs.destroy(self._created_org_name)
        all_orgs = self._api.admin_orgs.list()["data"]
        found_created_org = False
        for org in all_orgs:
            org_id = org["id"]
            if org_id == self._created_org_id:
                found_created_org = True
                break
        self.assertFalse(found_created_org)



"""
Module for Terraform Cloud API Endpoint: Notification Configurations.
"""

from .endpoint import TFCEndpoint

class TFCNotificationConfigurations(TFCEndpoint):
    """
    `Notification Configurations API Docs \
        <https://www.terraform.io/docs/cloud/api/notification-configurations.html>`_
    """
    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/notification-configurations"
        self._ws_base_url = f"{self._api_v2_base_url}/workspaces"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, workspace_id, payload):
        """
        ``POST /workspaces/:workspace_id/notification-configurations``

        `Notification Configurations Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/notification-configurations.html#create-a-notification-configuration>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/notification-configurations.html#sample-payload-for-generic-notification-configurations>`_
        """
        url = f"{self._ws_base_url}/{workspace_id}/notification-configurations"
        return self._create(url, payload)

    def list(self, workspace_id):
        """
        ``GET /workspaces/:workspace_id/notification-configurations``

        `Notification Configurations List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/notification-configurations.html#list-notification-configurations>`_
        """
        url = f"{self._ws_base_url}/{workspace_id}/notification-configurations"
        return self._list(url)

    def show(self, notification_config_id):
        """
        ``GET /notification-configurations/:notification-configuration-id``

        `Notification Configurations Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/notification-configurations.html#show-a-notification-configuration>`_
        """
        url = f"{self._endpoint_base_url}/{notification_config_id}"
        return self._show(url)

    def update(self, notification_config_id, payload):
        """
        ``PATCH /notification-configurations/:notification-configuration-id``

        `Notification Configurations Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/notification-configurations.html#update-a-notification-configuration>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/notification-configurations.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{notification_config_id}"
        return self._update(url, payload)

    def verify(self, notification_config_id):
        """
        ``POST /notification-configurations/:notification-configuration-id/actions/verify``

        `Notification Configurations Verify API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/notification-configurations.html#verify-a-notification-configuration>`_
        """
        url = f"{self._endpoint_base_url}/{notification_config_id}/actions/verify"
        return self._post(url)

    def destroy(self, notification_config_id):
        """
        ``DELETE /notification-configurations/:notification-configuration-id``

        `Notification Configurations Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/notification-configurations.html#delete-a-notification-configuration>`_
        """
        url = f"{self._endpoint_base_url}/{notification_config_id}"
        return self._destroy(url)



"""
Module for Terraform Cloud API Endpoint: State Version Outputs.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCStateVersionOutputs(TFCEndpoint):
    """
    `State Version Outputs API Docs \
        <https://www.terraform.io/docs/cloud/api/state-version-outputs.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/state-version-outputs"

    def required_entitlements(self):
        return [Entitlements.STATE_STORAGE]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def show(self, state_version_output_id):
        """
        ``GET /state-version-outputs/:state_version_output_id``

        `State Version Outputs Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/state-version-outputs.html#show-a-state-version-output>`_
        """
        url = f"{self._endpoint_base_url}/{state_version_output_id}"
        return self._show(url)



"""
Module for testing the Terraform Cloud API Endpoint: Admin Module Sharing.
"""

from .base import TestTFCBaseTestCase


class TestTFCAdminModuleSharing(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Admin Module Sharing.
    """

    _unittest_name = "adm-mod"
    _endpoint_being_tested = "admin_module_sharing"

    def setUp(self):
        # Create a temp org to manipulate in the test
        created_org = self._api.orgs.create(self._get_org_create_payload())["data"]
        self._created_org_name = created_org["attributes"]["name"]
        self._created_org_id = created_org["id"]
        self._created_org_external_id = created_org["attributes"]["external-id"]

    def tearDown(self):
        self._api.orgs.destroy(self._created_org_name)

    def test_admin_module_sharing(self):
        """
        Test the Admin Module Sharing API endpoints.
        """

        # Change the email address for the org, confirm the change.
        update_org_payload = {
            "data": {
                "type": "module-partnerships",
                "attributes": {
                    "module-consuming-organization-ids": [
                        self._created_org_external_id
                    ]
                }
            }
        }

        module_partnerships = \
            self._api.admin_module_sharing.update(self._test_org_name, update_org_payload)["data"]

        found_partnership = False
        for partnership in module_partnerships:
            if partnership["attributes"]["consuming-organization-id"] == \
                    self._created_org_external_id:
                found_partnership = True
                break

        self.assertTrue(found_partnership)



"""
Module for testing the Terraform Cloud API Endpoint: Org Tokens.
"""

from .base import TestTFCBaseTestCase


class TestTFCOrgTokens(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Org Tokens.
    """

    _unittest_name = "org-tok"
    _endpoint_being_tested = "org_tokens"

    def setUp(self):
        # Create a temp org to manipulate in the test.
        created_org = self._api.orgs.create(self._get_org_create_payload())["data"]
        self._created_org_name = created_org["attributes"]["name"]
        self._created_org_id = created_org["id"]

    def tearDown(self):
        self._api.orgs.destroy(self._created_org_name)

    def test_org_tokens(self):
        """
        Test the Org Tokens API endpoints.
        """

        self._api.set_org(self._created_org_id)

        # Create the token, confirm it was created
        created_token = self._api.org_tokens.create()["data"]
        created_token_id = created_token["id"]
        self.assertIsNotNone(created_token_id)

        # Remove the token, confirm it the request worked (no response)
        # since we don't have a way to look for them otherwise.
        removed_token_resp = self._api.org_tokens.destroy()
        self.assertIsNone(removed_token_resp)



"""
Module for testing the Terraform Cloud API Endpoint: Plan Exports.
"""

import time
import os

from terrasnek.exceptions import TFCHTTPNotFound
from .base import TestTFCBaseTestCase


class TestTFCPlanExports(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Plan Exports.
    """

    _unittest_name = "plan-exp"
    _endpoint_being_tested = "plan_exports"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        # Store the OAuth client ID to remove it at the end.
        oauth_client = self._api.oauth_clients.create(self._get_oauth_client_create_payload())
        self._oauth_client_id = oauth_client["data"]["id"]
        oauth_token_id = oauth_client["data"]["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a workspace using that token ID, save the workspace ID
        _ws_payload = self._get_ws_with_vcs_create_payload(oauth_token_id)
        workspace = self._api.workspaces.create(_ws_payload)["data"]
        self._ws_id = workspace["id"]

        # Configure the required variables on the workspace
        variable_payloads = [
            self._get_variable_create_payload(
                "email", self._test_email, self._ws_id),
            self._get_variable_create_payload(
                "org_name", self._test_org_name, self._ws_id),
            self._get_variable_create_payload(
                "TFE_TOKEN", self._test_api_token, self._ws_id, category="env", sensitive=True)
        ]
        for payload in variable_payloads:
            self._api.vars.create(payload)

        # Sleep for 3 seconds to give the WS time to create, takes a bit longer
        # on small TFE instances.
        time.sleep(3)

        # Start the run, store the run ID
        create_run_payload = self._get_run_create_payload(self._ws_id)
        self._run = self._api.runs.create(create_run_payload)["data"]
        self._run_id = self._run["id"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_plan_exports(self):
        """
        Test the Plan Exports API endpoints.
        """

        # Create a run and wait for the created run to complete it's plan
        created_run = self._api.runs.show(self._run_id)["data"]
        created_plan_id = created_run["relationships"]["plan"]["data"]["id"]

        # Timeout if the plan doesn't reach confirmable, this can happen
        # if the run is queued.
        created_run = self._created_run_timeout(self._run_id)

        # Export the plan, confirm the plan_id matches that in the exported plan
        create_plan_export_payload = {
            "data": {
                "type": "plan-exports",
                "attributes": {
                    "data-type": "sentinel-mock-bundle-v0"
                },
                "relationships": {
                    "plan": {
                        "data": {
                            "id": created_plan_id,
                            "type": "plans"
                        }
                    }
                }
            }
        }
        plan_export = self._api.plan_exports.create(create_plan_export_payload)["data"]
        self.assertEqual(
            created_plan_id, plan_export["relationships"]["plan"]["data"]["id"])

        # Verify that we get the right plan export back if we look it up by ID
        plan_export_id = plan_export["id"]
        shown_plan_export = self._api.plan_exports.show(plan_export_id)["data"]
        self.assertEqual(plan_export_id, shown_plan_export["id"])

        # If the download path already exists, clear it out, then download the plan
        # export, and verify it gets downloaded
        if os.path.exists(self._plan_export_tarball_target_path):
            os.remove(self._plan_export_tarball_target_path)
        self._api.plan_exports.download(
            plan_export_id, self._plan_export_tarball_target_path)
        self.assertTrue(os.path.exists(self._plan_export_tarball_target_path))
        os.remove(self._plan_export_tarball_target_path)

        # Destroy the plan export and confirm it's gone.
        self._api.plan_exports.destroy(plan_export_id)
        self.assertRaises(TFCHTTPNotFound, self._api.plan_exports.show, plan_export_id)



# -*- coding: utf-8 -*-
"""terrasnek Base Module

This module provides support for endpoints available in the Terraform Cloud
API. It's possible that some of these endpoints won't work if you're
working against an older version of Terraform Enterprise.

https://www.terraform.io/docs/cloud/api/index.html
"""

NAME = "terrasnek"



"""
Module for Terraform Cloud API Endpoint: Policy Checks.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCPolicyChecks(TFCEndpoint):
    """
    `Policy Checks API Docs \
        <https://www.terraform.io/docs/cloud/api/policy-checks.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/policy-checks"
        self._runs_api_v2_base_url = f"{self._api_v2_base_url}/runs"

    def required_entitlements(self):
        return [Entitlements.SENTINEL]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def show(self, policy_check_id):
        """
        ``GET /policy-checks/:id``

        `Policy Checks Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-checks.html#show-policy-check>`_
        """
        url = f"{self._endpoint_base_url}/{policy_check_id}"
        return self._show(url)

    def list(self, run_id):
        """
        ``GET /runs/:run_id/policy-checks``

        `Policy Checks List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-checks.html#list-policy-checks>`_
        """
        url = f"{self._runs_api_v2_base_url}/{run_id}/policy-checks"
        return self._list(url)

    def override(self, policy_check_id):
        """
        ``POST /policy-checks/:id/actions/override``

        `Policy Checks Override API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policy-checks.html#override-policy>`_
        """
        url = f"{self._endpoint_base_url}/{policy_check_id}/actions/override"
        return self._post(url)



"""
Module for Terraform Cloud API Endpoint: State Versions.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements, MAX_PAGE_SIZE

class TFCStateVersions(TFCEndpoint):
    """
    `State Versions API Docs \
        <https://www.terraform.io/docs/cloud/api/state-versions.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._state_version_api_v2_base_url = f"{self._api_v2_base_url}/state-versions"
        self._workspace_api_v2_base_url = f"{self._api_v2_base_url}/workspaces"

    def required_entitlements(self):
        return [Entitlements.STATE_STORAGE]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, workspace_id, payload):
        """
        ``POST /workspaces/:workspace_id/state-versions``

        `State Versions Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/state-versions.html#create-a-state-version>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/state-versions.html#sample-payload>`_
        """
        url = f"{self._workspace_api_v2_base_url}/{workspace_id}/state-versions"
        return self._create(url, payload)

    def get_current(self, workspace_id):
        """
        ``GET /workspaces/:workspace_id/current-state-version``

        `State Versions Get Current API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/state-versions.html#fetch-the-current-state-version-for-a-workspace>`_
        """
        url = f"{self._workspace_api_v2_base_url}/{workspace_id}/current-state-version"
        return self._get(url)

    def list(self, filters, page=None, page_size=None):
        """
        ``GET /state-versions``

        `State Versions List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/state-versions.html#list-state-versions-for-a-workspace>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/state-versions.html#query-parameters>`_):
            - ``filter[workspace][name]`` (Required)
            - ``filter[organization][name]`` (Required)
            - ``page`` (Optional)
            - ``page_size`` (Optional)

        Example filter(s):

        .. code-block:: python

            filters = [
                {
                    "keys": ["workspace", "name"],
                    "value": "foo"
                },
                {
                    "keys": ["organization", "name"],
                    "value": "bar"
                }
            ]
        """
        url = f"{self._state_version_api_v2_base_url}"
        return self._list(url, filters=filters, page=page, page_size=page_size)

    def list_all(self, filters):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every state version for a workspace.

        Returns an array of objects.
        """
        url = self._state_version_api_v2_base_url

        current_page_number = 1
        state_versions_resp = \
            self._list(url, filters=filters, page=current_page_number, page_size=MAX_PAGE_SIZE)
        total_pages = state_versions_resp["meta"]["pagination"]["total-pages"]

        state_versions = []
        while current_page_number <= total_pages:
            state_versions_resp = \
                self._list(url, filters=filters, page=current_page_number, page_size=MAX_PAGE_SIZE)
            state_versions += state_versions_resp["data"]
            current_page_number += 1

        return state_versions

    def show(self, state_version_id):
        """
        ``GET /state-versions/:state_version_id``

        `State Versions Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/state-versions.html#show-a-state-version>`_
        """
        url = f"{self._state_version_api_v2_base_url}/{state_version_id}"
        return self._show(url)



"""
Module for testing the Terraform Cloud API Endpoint: Policy Checks.
"""

import time

from .base import TestTFCBaseTestCase


class TestTFCPolicyChecks(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Policy Checks.
    """

    _unittest_name = "pol-chk"
    _endpoint_being_tested = "policy_checks"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        oauth_client = self._api.oauth_clients.create(self._get_oauth_client_create_payload())
        self._oauth_client_id = oauth_client["data"]["id"]
        oauth_token_id = oauth_client["data"]["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a workspace using that token ID, save the workspace ID
        ws_payload = self._get_ws_with_vcs_create_payload(oauth_token_id)
        workspace = self._api.workspaces.create(ws_payload)["data"]
        self._ws_id = workspace["id"]

        # Configure the required variables on the workspace
        variable_payloads = [
            self._get_variable_create_payload(
                "email", self._test_email, self._ws_id),
            self._get_variable_create_payload(
                "org_name", self._test_org_name, self._ws_id),
            self._get_variable_create_payload(
                "TFE_TOKEN", self._test_api_token, self._ws_id, category="env", sensitive=True)
        ]
        for payload in variable_payloads:
            self._api.vars.create(payload)

        # Sleep for 1 second to give the WS time to create
        time.sleep(1)

        # Add the policy set
        # Add a policy set to TFC
        pol_set_create_payload = self._get_policy_set_create_payload(oauth_token_id)
        created_policy_set = self._api.policy_sets.create(pol_set_create_payload)["data"]
        self._created_policy_set_id = created_policy_set["id"]

        # Attach the policy set to the workspace we created, confirm it's attached
        attach_to_workspace = {
            "data": [
                {
                    "id": self._ws_id, "type": "workspaces"
                },
            ]
        }
        self._api.policy_sets.attach_policy_set_to_workspaces(\
            self._created_policy_set_id, attach_to_workspace)

        # Wait a second to make sure the policy is attached before creating the run
        time.sleep(1)

        # Start the run, store the run ID
        create_run_payload = self._get_run_create_payload(self._ws_id)
        self._run = self._api.runs.create(create_run_payload)["data"]
        self._run_id = self._run["id"]


    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)
        self._api.policy_sets.destroy(self._created_policy_set_id)

    def test_policy_checks(self):
        """
        Test the Policy Checks API endpoints.
        """

        # Create a run and wait for the created run to complete it's plan
        created_run = self._api.runs.show(self._run_id)["data"]
        while created_run["attributes"]["status"] != "policy_override":
            self._logger.debug("Waiting for created run to get to policy override status...")
            created_run = self._api.runs.show(self._run_id)["data"]
            time.sleep(1)
        self._logger.debug("Plan successful.")

        run_id = created_run["id"]

        # List the policy checks, make sure we only have one and that it failed
        pol_checks = self._api.policy_checks.list(run_id)["data"]
        self.assertEqual(len(pol_checks), 1)
        self.assertFalse(pol_checks[0]["attributes"]["result"]["result"])

        pol_check_id = pol_checks[0]["id"]
        shown_pol_check = self._api.policy_checks.show(pol_check_id)["data"]
        self.assertEqual(pol_check_id, shown_pol_check["id"])

        # Override the policy check that soft failed, confirm we can continue
        self._api.policy_checks.override(pol_check_id)

        created_run = self._api.runs.show(self._run_id)["data"]
        self.assertTrue(created_run["attributes"]["actions"]["is-confirmable"])



"""
Module for Terraform Cloud API Endpoint: Variables.
"""

from .endpoint import TFCEndpoint

class TFCVars(TFCEndpoint):
    """
    `Vars API Docs \
        <https://www.terraform.io/docs/cloud/api/variables.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/vars"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST /vars``

        `Vars Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variables.html#create-a-variable>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/variables.html#sample-payload>`_
        """
        return self._create(self._endpoint_base_url, payload)

    def list(self, workspace_name=None):
        """
        ``GET /vars``

        `Vars List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variables.html#list-variables>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/variables.html#query-parameters>`_):
            - ``filter[workspace][name]`` (Required)
        """
        url = f"{self._endpoint_base_url}?filter[organization][name]={self._org_name}"

        if workspace_name is not None:
            url += f"&filter[workspace][name]={workspace_name}"

        return self._list(url)

    def update(self, variable_id, payload):
        """
        ``PATCH /vars/:variable_id``

        `Vars Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variables.html#update-variables>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/variables.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{variable_id}"
        return self._update(url, payload)

    def destroy(self, variable_id):
        """
        ``DELETE /vars/:variable_id``

        `Vars Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/variables.html#delete-variables>`_
        """
        url = f"{self._endpoint_base_url}/{variable_id}"
        return self._destroy(url)



"""
Module for testing the Terraform Cloud API Endpoint: OAuth Tokens.
"""

from .base import TestTFCBaseTestCase


class TestTFCOAuthTokens(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: OAuth Tokens.
    """

    _unittest_name = "oa-tokens"
    _endpoint_being_tested = "oauth_tokens"

    def setUp(self):
        self._oauth_client = self._api.oauth_clients.create(\
            self._get_oauth_client_create_payload())["data"]
        self._oauth_client_id = self._oauth_client["id"]

    def tearDown(self):
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_oauth_tokens(self):
        """
        Test the OAuth Tokens API endpoints.
        """

        # List all the tokens and make sure there is one associated to your client
        oauth_tokens = self._api.oauth_tokens.list(
            self._oauth_client_id)["data"]
        self.assertEqual(len(oauth_tokens), 1)

        # Make sure the OAuth client parent is as expected
        parent_oauth_client_id = oauth_tokens[0]["relationships"]["oauth-client"]["data"]["id"]
        self.assertEqual(parent_oauth_client_id, self._oauth_client_id)

        # Look up the token by it's ID and make sure the results are as expected
        oauth_token_id = oauth_tokens[0]["id"]
        shown_oauth_token = self._api.oauth_tokens.show(oauth_token_id)
        self.assertEqual(oauth_token_id, shown_oauth_token["data"]["id"])

        # Create an empty update payload (it requires a real Git SSH key), so publish
        # an update and verify the response type is what we expect
        update_payload = {
            "data": {
                "id": oauth_token_id,
                "type": "oauth-tokens",
                "attributes": {
                }
            }
        }
        updated_oauth_token = self._api.oauth_tokens.update(oauth_token_id, update_payload)["data"]
        self.assertEqual("oauth-tokens", updated_oauth_token["type"])

        # Delete the OAuth token and confirm it's no longer present
        self._api.oauth_tokens.destroy(oauth_token_id)
        oauth_tokens = self._api.oauth_tokens.list(
            self._oauth_client_id)["data"]

        found_token = False
        for oauth_token in oauth_tokens:
            if oauth_token_id == oauth_token["id"]:
                found_token = True
                break
        self.assertFalse(found_token)



"""
Module for testing the Terraform Cloud API Endpoint: State Versions.
"""

from .base import TestTFCBaseTestCase


class TestTFCStateVersions(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: State Versions.
    """

    _unittest_name = "state-ver"
    _endpoint_being_tested = "state_versions"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        # Store the OAuth client ID to remove it at the end.
        self._oauth_client = self._api.oauth_clients.create(self._get_oauth_client_create_payload())
        self._oauth_client_id = self._oauth_client["data"]["id"]
        oauth_token_id = \
            self._oauth_client["data"]["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a workspace using that token ID, save the workspace ID
        _ws_create_with_vcs_payload = self._get_ws_with_vcs_create_payload(oauth_token_id)
        self._ws = self._api.workspaces.create(_ws_create_with_vcs_payload)
        self._ws_id = self._ws["data"]["id"]
        self._ws_name = self._ws["data"]["attributes"]["name"]

        # Create an example config version for the unittest
        self._config_version = self._api.config_versions.create(
            self._ws_id, self._get_config_version_create_payload())["data"]
        self._config_version_upload_url = self._config_version["attributes"]["upload-url"]
        self._cv_id = self._config_version["id"]
        self._api.config_versions.upload(
            self._config_version_upload_tarball_path, self._config_version_upload_url)

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_state_versions(self):
        """
        Test the State Version API endpoints.
        """

        # List the existing state versions, confirm there are none.
        test_filters = [
            {
                "keys": ["workspace", "name"],
                "value": self._ws_name
            },
            {
                "keys": ["organization", "name"],
                "value": self._test_org_name
            }
        ]
        state_versions = self._api.state_versions.list(\
            filters=test_filters, page=0, page_size=50)["data"]
        self.assertEqual(len(state_versions), 0)
        self._api.workspaces.lock(self._ws_id, {"reason": "Unit testing."})

        # Create a state version, confirm we now have one
        self._api.state_versions.create(
            self._ws_id, self._get_state_version_create_payload())
        self._api.workspaces.unlock(self._ws_id)
        state_versions = self._api.state_versions.list(filters=test_filters)["data"]
        self.assertNotEqual(len(state_versions), 0)

        all_state_versions = self._api.state_versions.list_all(filters=test_filters)
        self.assertNotEqual(len(all_state_versions), 0)

        # Get the most current state version, confirm it matches the one we created
        state_version = state_versions[0]
        sv_id = state_version["id"]
        current_state_version = self._api.state_versions.get_current(self._ws_id)[
            "data"]
        self.assertEqual(sv_id, current_state_version["id"])

        # Show the state version by ID, confirm that they match
        shown_state_version = self._api.state_versions.show(sv_id)["data"]
        self.assertEqual(sv_id, shown_state_version["id"])



"""
Module for Terraform Cloud API Endpoint: Agent Pools and Agents.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCAgents(TFCEndpoint):
    """
    `Agents API Docs \
        <https://www.terraform.io/docs/cloud/api/agents.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._agent_pools_api_v2_base_url = f"{self._api_v2_base_url}/agent-pools"
        self._agents_api_v2_base_url = f"{self._api_v2_base_url}/agents"
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations"

    def required_entitlements(self):
        return [Entitlements.AGENTS]

    def terraform_cloud_only(self):
        return True

    def terraform_enterprise_only(self):
        return False

    def create_pool(self, payload):
        """
        ``POST /organizations/:organization_name/agent-pool``

        `Agents Create Pool API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agents.html#create-an-agent-pool>`_
        """
        url = f"{self._org_api_v2_base_url}/{self._org_name}/agent-pools"
        return self._create(url, payload)

    def list_pools(self):
        """
        ``GET /organizations/:organization_name/agent-pools``

        `Agents List Pools API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agents.html#list-agent-pools>`_
        """
        url = f"{self._org_api_v2_base_url}/{self._org_name}/agent-pools"
        return self._list(url)

    def list(self, agent_pool_id, filters=None):
        """
        ``GET /agent-pools/:agent_pool_id/agents``

        `Agents List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agents.html#list-agents>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/agents.html#query-parameters>`_):
            - ``filter[last-ping-since]`` (Optional)

        Example filter(s):

        .. code-block:: python

            filters = [
                {
                    "keys": ["last-ping-since"],
                    "value": "foo"
                }
            ]
        """
        url = f"{self._agent_pools_api_v2_base_url}/{agent_pool_id}/agents"
        return self._list(url, filters=filters)

    def show_pool(self, agent_pool_id):
        """
        ``GET /agent-pools/:id``

        `Agents Show Pool API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agents.html#show-an-agent-pool>`_
        """
        url = f"{self._agent_pools_api_v2_base_url}/{agent_pool_id}"
        return self._show(url)

    def show(self, agent_id):
        """
        ``GET /agents/:id``

        `Agents Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agents.html#show-an-agent>`_
        """
        url = f"{self._agents_api_v2_base_url}/{agent_id}"
        return self._show(url)

    def update(self, agent_pool_id, payload):
        """
        ``PATCH /agent-pools/:id``

        `Agents Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agents.html#update-an-agent-pool>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/agents.html#sample-payload-1>`_
        """
        url = f"{self._agent_pools_api_v2_base_url}/{agent_pool_id}"
        return self._update(url, payload)

    def destroy(self, agent_pool_id):
        """
        ``DELETE /agent-pools/:agent_pool_id``

        `Agents Destory Pool API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/agents.html#delete-an-agent-pool>`_
        """
        url = f"{self._agent_pools_api_v2_base_url}/{agent_pool_id}"
        return self._delete(url)



"""
Module for testing the Terraform Cloud API Endpoint: Cost Estimates.
"""

import time

from .base import TestTFCBaseTestCase

from ._constants import \
    AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY

class TestTFCCostEstimates(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Cost Estimates.
    """

    _unittest_name = "cst-est"
    _endpoint_being_tested = "cost_estimates"

    def setUp(self):
        # Create an OAuth client for the test and extract it's the token ID
        # Store the OAuth client ID to remove it at the end.
        oauth_client_payload = self._get_oauth_client_create_payload()
        oauth_client = self._api.oauth_clients.create(oauth_client_payload)
        self._oauth_client_id = oauth_client["data"]["id"]
        oauth_token_id = oauth_client["data"]["relationships"]["oauth-tokens"]["data"][0]["id"]

        # Create a workspace using that token ID, save the workspace ID
        _ws_payload = self._get_ws_with_vcs_create_payload(oauth_token_id, working_dir="aws")
        workspace = self._api.workspaces.create(_ws_payload)["data"]
        self._ws_id = workspace["id"]

        # Configure the required variables on the workspace
        variable_payloads = [
            self._get_variable_create_payload(
                "AWS_ACCESS_KEY_ID", AWS_ACCESS_KEY_ID, \
                self._ws_id, category="env", sensitive=True),
            self._get_variable_create_payload(
                "AWS_SECRET_ACCESS_KEY", AWS_SECRET_ACCESS_KEY, \
                self._ws_id, category="env", sensitive=True)
        ]
        for payload in variable_payloads:
            self._api.vars.create(payload)

        # Sleep for 1 second to give the WS time to create
        time.sleep(1)

        # Enable cost estimation if this is TFE
        if not self._api.is_terraform_cloud():
            # Configure the AWS credentials for it at the admin level
            update_payload = {
                "data": {
                    "attributes": {
                        "enabled": True,
                        "aws-enabled": True,
                        "aws-access-key-id": AWS_ACCESS_KEY_ID,
                        "aws-secret-key": AWS_SECRET_ACCESS_KEY
                    }
                }
            }
            updated_admin_cost_est_settings = \
                self._api.admin_settings.update_cost_estimation(update_payload)["data"]
            self.assertTrue(updated_admin_cost_est_settings["attributes"]["enabled"])

        # Enable cost estimation on the org level
        update_payload = {
            "data": {
                "attributes": {
                    "cost-estimation-enabled": True
                }
            }
        }
        self._api.orgs.update(self._test_org_name, update_payload)

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)
        self._api.oauth_clients.destroy(self._oauth_client_id)

    def test_cost_estimates(self):
        """
        Test the Cost Estimates API endpoints.
        """

        # Create a run
        create_run_payload = self._get_run_create_payload(self._ws_id)
        run = self._api.runs.create(create_run_payload)["data"]
        run_id = run["id"]

        # Timeout if the plan doesn't reach confirmable, this can happen
        # if the run is queued.
        created_run = self._created_run_timeout(run_id)

        # Get the Cost Estimate ID from the plan, and confirm that we can show it
        cost_estimate_id = created_run["relationships"]["cost-estimate"]["data"]["id"]
        cost_estimate = self._api.cost_estimates.show(cost_estimate_id)["data"]
        self.assertEqual("cost-estimates", cost_estimate["type"])



"""
Module for Terraform Cloud API Endpoint: Policies.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements, MAX_PAGE_SIZE

class TFCPolicies(TFCEndpoint):
    """
    `Policies API Docs \
        <https://www.terraform.io/docs/cloud/api/policies.html>`_
    """
    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/policies"
        self._org_api_v2_base_url = f"{self._api_v2_base_url}/organizations/{org_name}/policies"

    def required_entitlements(self):
        return [Entitlements.SENTINEL]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST /organizations/:organization_name/policies``

        `Policies Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policies.html#create-a-policy>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policies.html#sample-payload>`_
        """
        return self._create(self._org_api_v2_base_url, payload)

    def list(self, page=None, page_size=None, search=None):
        """
        ``GET /organizations/:organization_name/policies``

        `Policies List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policies.html#list-policies>`_

        Query Parameter(s) (`details \
            <https://www.terraform.io/docs/cloud/api/policies.html#query-parameters>`_):
            - ``page`` (Optional)
            - ``page_size`` (Optional)
            - ``search`` (Optional)
        """
        return self._list(\
            self._org_api_v2_base_url, page=page, page_size=page_size, search=search)

    def list_all(self, search=None):
        """
        This function does not correlate to an endpoint in the TFC API Docs specifically,
        but rather is a helper function to wrap the `list` endpoint, which enumerates out
        every page so users do not have to implement the paging logic every time they just
        want to list every policy for an organization.

        Returns an array of objects.
        """
        current_page_number = 1
        policies_resp = \
            self.list(page=current_page_number, page_size=MAX_PAGE_SIZE, search=search)
        total_pages = policies_resp["meta"]["pagination"]["total-pages"]

        policies = []
        while current_page_number <= total_pages:
            policies_resp = \
                self.list(page=current_page_number, page_size=MAX_PAGE_SIZE, search=search)
            policies += policies_resp["data"]
            current_page_number += 1

        return policies

    def show(self, policy_id):
        """
        ``GET /policies/:policy_id``

        `Policies Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policies.html#show-a-policy>`_
        """
        url = f"{self._endpoint_base_url}/{policy_id}"
        return self._show(url)

    def update(self, policy_id, payload):
        """
        ``PATCH /policies/:policy_id``

        `Policies Update API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policies.html#update-a-policy>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policies.html#sample-payload-2>`_
        """
        url = f"{self._endpoint_base_url}/{policy_id}"
        return self._update(url, payload)

    def get_policy_text(self, policy_id):
        """
        ``GET /policies/:policy_id/download``

        This endpoint is currently not documented in the offical TFC API docs. You can find a
        reference to it in the `sample response \
            <https://www.terraform.io/docs/cloud/api/policies.html#sample-response-1>` for the
        `show` function.
        """
        url = f"{self._endpoint_base_url}/{policy_id}/download"
        byte_results = self._get(url, return_raw=True, allow_redirects=True)
        return byte_results.decode("utf-8")

    def upload(self, policy_id, payload):
        """
        ``PUT /policies/:policy_id/upload``

        `Policies Upload API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policies.html#upload-a-policy>`_

        `Upload Sample Payload \
            <https://www.terraform.io/docs/cloud/api/policies.html#sample-payload-1>`_
        """
        url = f"{self._endpoint_base_url}/{policy_id}/upload"
        return self._put(url, octet=True, data=payload)

    def destroy(self, policy_id):
        """
        ``DELETE /policies/:policy_id``

        `Policies Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/policies.html#delete-a-policy>`_
        """
        url = f"{self._endpoint_base_url}/{policy_id}"
        return self._destroy(url)



"""
Module for Terraform Cloud API Endpoint: Plans.
"""

from .endpoint import TFCEndpoint
from ._constants import Entitlements

class TFCPlans(TFCEndpoint):
    """
    `Plans API Docs \
        <https://www.terraform.io/docs/cloud/api/plans.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/plans"
        self._runs_base_url = f"{self._api_v2_base_url}/runs"

    def required_entitlements(self):
        return [Entitlements.OPERATIONS]

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def show(self, plan_id):
        """
        ``GET /plans/:id``

        `Plans Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/plans.html#show-a-plan>`_
        """
        url = f"{self._endpoint_base_url}/{plan_id}"
        return self._show(url)

    def download_json(self, target_path, plan_id=None, run_id=None):
        """
        ``GET /plans/:id/json-output``
        ``GET /runs/:id/plan/json-output``

        `Plans Download JSON API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/plans.html#retrieve-the-json-execution-plan>`_
        """
        if plan_id is not None:
            url = f"{self._endpoint_base_url}/{plan_id}/json-output"
        elif run_id is not None:
            url = f"{self._runs_base_url}/{run_id}/plan/json-output"
        else:
            self._logger.error("Arguments plan_id or run_id must be defined")

        results = self._get(url, return_raw=True, allow_redirects=True)

        with open(target_path, 'wb') as target_file:
            target_file.write(results)



"""
Module for Terraform Cloud API Endpoint: Plan Exports.
"""

from .endpoint import TFCEndpoint

class TFCPlanExports(TFCEndpoint):
    """
    `Plan Exports API Docs \
        <https://www.terraform.io/docs/cloud/api/plan-exports.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/plan-exports"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self, payload):
        """
        ``POST /plan-exports``

        `Plan Exports Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/plan-exports.html#create-a-plan-export>`_

        `Create Sample Payload \
            <https://www.terraform.io/docs/cloud/api/plan-exports.html#sample-payload>`_
        """

        return self._create(self._endpoint_base_url, payload)

    def show(self, plan_export_id):
        """
        ``GET /plan-exports/:id``

        `Plan Exports Show API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/plan-exports.html#show-a-plan-export>`_
        """
        url = f"{self._endpoint_base_url}/{plan_export_id}"
        return self._show(url)

    def download(self, plan_export_id, target_path):
        """
        ``GET /plan-exports/:id/download``

        `Plan Exports Download API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/plan-exports.html#download-exported-plan-data>`_
        """
        url = f"{self._endpoint_base_url}/{plan_export_id}/download"
        results = self._get(url, return_raw=True, allow_redirects=True)
        with open(target_path, 'wb') as target_file:
            target_file.write(results)

    def destroy(self, plan_export_id):
        """
        ``DELETE /plan-exports/:id``

        `Plan Exports Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/plan-exports.html#delete-exported-plan-data>`_
        """
        url = f"{self._endpoint_base_url}/{plan_export_id}"
        return self._destroy(url)



"""
Module for Terraform Cloud API Endpoint: Account.
"""

from .endpoint import TFCEndpoint

class TFCAccount(TFCEndpoint):
    """
    `Account API Docs \
        <https://www.terraform.io/docs/cloud/api/account.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._api_v2_base_url}/account"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def show(self):
        """
        ``GET /account/details``

        `Show Account API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/account.html#get-your-account-details>`_
        """
        url = f"{self._endpoint_base_url}/details"
        return self._show(url)

    def update(self, payload):
        """
        ``PATCH /account/update``

        `Update Account API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/account.html#update-your-account-info>`_

        `Update Sample Payload \
            <https://www.terraform.io/docs/cloud/api/account.html#sample-payload>`_
        """
        url = f"{self._endpoint_base_url}/update"
        return self._patch(url, payload)

    def change_password(self, data):
        """
        ``PATCH /account/password``

        `Change Account Password API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/account.html#change-your-password>`_

        `Change Password Sample Payload \
            <https://www.terraform.io/docs/cloud/api/account.html#sample-payload>`_
        """
        url = f"{self._endpoint_base_url}/password"
        return self._patch(url, data)



"""
Module for testing the Terraform Cloud API Endpoint: Teams.
"""

from .base import TestTFCBaseTestCase


class TestTFCTeams(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Teams.
    """

    _unittest_name = "teams"
    _endpoint_being_tested = "teams"

    def test_teams(self):
        """
        Test the Teams API endpoints.
        """

        # List all the teams, confirm that the response type
        teams = self._api.teams.list()["data"]
        self.assertEqual("teams", teams[0]["type"])

        # Create a new team, confirm that it has been created
        new_team = self._api.teams.create(
            self._get_team_create_payload())["data"]
        new_team_id = new_team["id"]
        all_teams = self._api.teams.list()["data"]
        found_team = False
        for team in all_teams:
            if team["id"] == new_team_id:
                found_team = True
                break
        self.assertTrue(found_team)

        # Show the newly created team, assert that the response matches the ID we fed in.
        shown_team = self._api.teams.show(new_team_id)["data"]
        self.assertEqual(shown_team["id"], new_team_id)

        # Update the team to have VCS management access, confirm the changes took effect.
        update_payload = {
            "data": {
                "type": "teams",
                "attributes": {
                    "visibilty": "organization",
                    "organization-access": {
                        "manage-vcs-settings": True
                    }
                }
            }
        }
        updated_team = self._api.teams.update(new_team_id, update_payload)["data"]
        self.assertTrue(updated_team["attributes"]["organization-access"]["manage-vcs-settings"])

        # Destroy the team, confirm it's gone
        self._api.teams.destroy(new_team_id)
        all_teams = self._api.teams.list()["data"]
        found_team = False
        for team in all_teams:
            if team["id"] == new_team_id:
                found_team = True
                break
        self.assertFalse(found_team)



"""
Module for testing the Terraform Cloud API Endpoint: Config Versions.
"""

from .base import TestTFCBaseTestCase


class TestTFCConfigVersions(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Config Versions.
    """

    _unittest_name = "cnf-ver"
    _endpoint_being_tested = "config_versions"

    def setUp(self):
        self._ws = self._api.workspaces.create(self._get_ws_without_vcs_create_payload())
        self._ws_id = self._ws["data"]["id"]

    def tearDown(self):
        self._api.workspaces.destroy(workspace_id=self._ws_id)

    def test_config_versions(self):
        """
        Test the Config Versions API endpoints.
        """

        # two upload methods to test; use a different 1st positional argument for each test
        upload_tests = (
            (self._api.config_versions.upload, self._config_version_upload_tarball_path),
            (self._api.config_versions.upload_from_string, self._config_version_upload_string),
        )
        for upload_handle, source in upload_tests:

            with self.subTest():

                # Create a new config version
                config_version = self._api.config_versions.create(
                    self._ws_id, self._get_config_version_create_payload())["data"]
                cv_id = config_version["id"]

                # List all of the config versions for the workspace
                config_versions = self._api.config_versions.list(self._ws_id)["data"]

                # Confirm we found the newly created config version
                found_conf_ver = False
                for conf_ver in config_versions:
                    if cv_id == conf_ver["id"]:
                        found_conf_ver = True
                        break
                self.assertTrue(found_conf_ver)

                # Confirm the config version status is "pending" or "uploaded" as well
                uploaded_or_pending = \
                    config_versions[0]["attributes"]["status"] in ["pending", "uploaded"]
                self.assertTrue(uploaded_or_pending)

                # Test the show method on that same config version ID
                shown_config_version = self._api.config_versions.show(cv_id)["data"]

                # Confirm the results match the same ID we looked up
                self.assertEqual(shown_config_version["id"], cv_id)

                # Upload the .tf code and confirm it's been uploaded
                upload_url = shown_config_version["attributes"]["upload-url"]
                upload_handle(source, upload_url)

                config_versions = self._api.config_versions.list(self._ws_id)["data"]
                uploaded_or_pending = \
                    config_versions[0]["attributes"]["status"] in ["pending", "uploaded"]
                self.assertTrue(uploaded_or_pending)

                all_config_versions = self._api.config_versions.list_all(self._ws_id)
                found_conf_ver = False
                for conf_ver in all_config_versions:
                    if cv_id == conf_ver["id"]:
                        found_conf_ver = True
                        break
                self.assertTrue(found_conf_ver)



"""
Module for testing the Terraform Cloud API Endpoint: Admin Settings.
"""

import random

from .base import TestTFCBaseTestCase


class TestTFCAdminSettings(TestTFCBaseTestCase):
    """
    Class for testing the Terraform Cloud API Endpoint: Admin Settings.
    """

    _unittest_name = "admin-settings"
    _endpoint_being_tested = "admin_settings"

    def test_admin_settings_general(self):
        """
        Test the Admin Settings General API endpoints.
        """
        # List all the general settings, verify the response type
        all_general = self._api.admin_settings.list_general()["data"]
        self.assertEqual("general-settings", all_general["type"])

        # Update the rate limit to a random number, verify it worked.
        updated_rate_limit = random.randint(45, 60)
        update_payload = {
            "data": {
                "attributes": {
                    "api-rate-limiting-enabled": True,
                    "api-rate-limit": updated_rate_limit
                }
            }
        }
        updated_settings = self._api.admin_settings.update_general(update_payload)["data"]
        self.assertEqual(updated_rate_limit, updated_settings["attributes"]["api-rate-limit"])

    def test_admin_settings_cost_estimation(self):
        """
        Test the Admin Settings Cost Estimation API endpoints.
        """

        # List all the cost estimation settings, verify the response type
        all_cost_est = self._api.admin_settings.list_cost_estimation()["data"]
        self.assertEqual("cost-estimation-settings", all_cost_est["type"])

        # Update the cost estimation settings to enable cost estimation,
        # verify that it's been enabled
        update_payload = {
            "data": {
                "attributes": {
                    "enabled": True,
                    "aws-enabled": True,
                    "aws-access-key-id": "foo",
                    "aws-secret-key": "bar",
                    "azure-enabled": False,
                    "gcp-enabled": False
                }
            }
        }
        updated_cost_est = self._api.admin_settings.update_cost_estimation(update_payload)["data"]
        self.assertTrue(updated_cost_est["attributes"]["enabled"])
        self.assertTrue(updated_cost_est["attributes"]["aws-enabled"])

    def test_admin_settings_saml(self):
        """
        Test the Admin Settings SAML API endpoints.
        """

        # List all the SAML settings, verify the response type
        all_saml = self._api.admin_settings.list_saml()["data"]
        self.assertEqual("saml-settings", all_saml["type"])

        # Update the SAML settings to enable SAML, confirm it's enabled
        update_payload = {
            "data": {
                "attributes": {
                    "enabled": True,
                    "debug": False,
                    "idp-cert": "NEW-CERTIFICATE",
                    "slo-endpoint-url": "https://example.com/slo",
                    "sso-endpoint-url": "https://example.com/sso",
                    "attr-username": "Username",
                    "attr-groups": "MemberOf",
                    "attr-site-admin": "SiteAdmin",
                    "site-admin-role": "site-admins",
                    "sso-api-token-session-timeout": 1209600
                }
            }
        }
        updated_saml = self._api.admin_settings.update_saml(update_payload)["data"]
        self.assertTrue(updated_saml["attributes"]["enabled"])

        # TODO: revoke_previous_saml_idp_cert

        # Disable SAML after running the test so that it's easier to log in with
        # username and password.
        update_payload["data"]["attributes"]["enabled"] = False
        updated_saml = self._api.admin_settings.update_saml(update_payload)["data"]
        self.assertFalse(updated_saml["attributes"]["enabled"])

    def test_admin_settings_smtp(self):
        """
        Test the Admin Settings SMTP API endpoints.
        """

        # List all the SMTP settings, verify the response type
        all_smtp = self._api.admin_settings.list_smtp()["data"]
        self.assertEqual("smtp-settings", all_smtp["type"])

        # TODO: need to use a real SMTP server to finish this test.

    def test_admin_settings_twilio(self):
        """
        Test the Admin Settings Twilio API endpoints.
        """

        # List all the Twilio settings, verify the response type
        all_twilio = self._api.admin_settings.list_twilio()["data"]
        self.assertEqual("twilio-settings", all_twilio["type"])

        # TODO: need to use a real twilio account to finish this test.

    def test_admin_settings_customization(self):
        """
        Test the Admin Settings Customization API endpoints.
        """

        # List all the customization settings, verify the response type
        all_customization = self._api.admin_settings.list_customization()["data"]
        self.assertEqual("customization-settings", all_customization["type"])

        # Update the support email, confirm it works
        email_to_update_to = "foo@bar.com"
        update_payload = {
            "data": {
                "attributes": {
                    "support-email-address": email_to_update_to
                }
            }
        }
        updated_customization = \
            self._api.admin_settings.update_customization(update_payload)["data"]
        self.assertEqual(\
            email_to_update_to, updated_customization["attributes"]["support-email-address"])



"""
Module for Terraform Cloud API Endpoint: Org Tokens.
"""

from .endpoint import TFCEndpoint

class TFCOrgTokens(TFCEndpoint):
    """
    `Org Tokens API Docs \
        <https://www.terraform.io/docs/cloud/api/organization-tokens.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = \
            f"{self._api_v2_base_url}/organizations/{org_name}/authentication-token"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return False

    def terraform_enterprise_only(self):
        return False

    def create(self):
        """
        ``POST /organizations/:organization_name/authentication-token``

        `Org Tokens Create API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-tokens.html#generate-a-new-organization-token>`_
        """
        return self._create(self._endpoint_base_url, None)

    def destroy(self):
        """
        ``DELETE /organizations/:organization/authentication-token``

        `Org Tokens Destroy API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/organization-tokens.html#delete-the-organization-token>`_
        """
        url = f"{self._endpoint_base_url}"
        return self._destroy(url)



"""
Module for container class of all TFC endpoints and high level exceptions around
API access.
"""

import json
import logging
import requests
import urllib3

from ._constants import TFC_SAAS_URL, TFC_SAAS_HOSTNAME, HTTP_OK, API_LOG_LEVEL
from .exceptions import TFCHTTPNotFound

from .account import TFCAccount
from .admin_module_sharing import TFCAdminModuleSharing
from .admin_orgs import TFCAdminOrgs
from .admin_runs import TFCAdminRuns
from .admin_settings import TFCAdminSettings
from .admin_terraform_versions import TFCAdminTerraformVersions
from .admin_users import TFCAdminUsers
from .admin_workspaces import TFCAdminWorkspaces
from .agents import TFCAgents
from .agent_tokens import TFCAgentTokens
from .applies import TFCApplies
from .audit_trails import TFCAuditTrails
from .config_versions import TFCConfigVersions
from .cost_estimates import TFCCostEstimates
from .feature_sets import TFCFeatureSets
from .ip_ranges import TFCIPRanges
from .invoices import TFCInvoices
from .oauth_clients import TFCOAuthClients
from .oauth_tokens import TFCOAuthTokens
from .orgs import TFCOrgs
from .org_memberships import TFCOrgMemberships
from .org_tokens import TFCOrgTokens
from .plans import TFCPlans
from .plan_exports import TFCPlanExports
from .policies import TFCPolicies
from .policy_checks import TFCPolicyChecks
from .policy_sets import TFCPolicySets
from .policy_set_params import TFCPolicySetParams
from .notification_configs import TFCNotificationConfigurations
from .registry_modules import TFCRegistryModules
from .run_triggers import TFCRunTriggers
from .runs import TFCRuns
from .state_versions import TFCStateVersions
from .state_version_outputs import TFCStateVersionOutputs
from .ssh_keys import TFCSSHKeys
from .subscriptions import TFCSubscriptions
from .teams import TFCTeams
from .team_access import TFCTeamAccess
from .team_memberships import TFCTeamMemberships
from .team_tokens import TFCTeamTokens
from .users import TFCUsers
from .user_tokens import TFCUserTokens
from .vars import TFCVars
from .workspace_vars import TFCWorkspaceVars
from .workspaces import TFCWorkspaces

# Suppress insecure TLS warnings
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class InvalidTFCTokenException(Exception):
    """Cannot instantiate TFC API class without a valid TFC_TOKEN."""


class TFC():
    """
    Super class for access to all TFC Endpoints.
    """

    # This dict specifies which class should be used for each attribute of this API class,
    # which simplifies the initialization of each endpoint since they share the same initialization
    # values.
    _class_for_attr_dict = {
        "org-not-required": {
            "admin_orgs": TFCAdminOrgs,
            "admin_module_sharing": TFCAdminModuleSharing,
            "admin_runs": TFCAdminRuns,
            "admin_settings": TFCAdminSettings,
            "admin_terraform_versions": TFCAdminTerraformVersions,
            "admin_users": TFCAdminUsers,
            "admin_workspaces": TFCAdminWorkspaces,
            "audit_trails": TFCAuditTrails,
            "ip_ranges": TFCIPRanges,
            "orgs": TFCOrgs
        },
        "org-required": {
            "account": TFCAccount,
            "applies": TFCApplies,
            "agents": TFCAgents,
            "agent_tokens": TFCAgentTokens,
            "config_versions": TFCConfigVersions,
            "cost_estimates": TFCCostEstimates,
            "feature_sets": TFCFeatureSets,
            "invoices": TFCInvoices,
            "oauth_clients": TFCOAuthClients,
            "oauth_tokens": TFCOAuthTokens,
            "org_memberships": TFCOrgMemberships,
            "org_tokens": TFCOrgTokens,
            "plans": TFCPlans,
            "plan_exports": TFCPlanExports,
            "policies": TFCPolicies,
            "policy_checks": TFCPolicyChecks,
            "policy_sets": TFCPolicySets,
            "policy_set_params": TFCPolicySetParams,
            "notification_configs": TFCNotificationConfigurations,
            "registry_modules": TFCRegistryModules,
            "run_triggers": TFCRunTriggers,
            "runs": TFCRuns,
            "ssh_keys": TFCSSHKeys,
            "state_versions": TFCStateVersions,
            "state_version_outputs": TFCStateVersionOutputs,
            "subscriptions": TFCSubscriptions,
            "teams": TFCTeams,
            "team_access": TFCTeamAccess,
            "team_memberships": TFCTeamMemberships,
            "team_tokens": TFCTeamTokens,
            "users": TFCUsers,
            "user_tokens": TFCUserTokens,
            "vars": TFCVars,
            "workspace_vars": TFCWorkspaceVars,
            "workspaces": TFCWorkspaces
        }
    }

    def __init__(self, api_token, url=TFC_SAAS_URL, verify=True, log_level=API_LOG_LEVEL):
        if api_token is None:
            raise InvalidTFCTokenException

        self._logger = logging.getLogger(self.__class__.__name__)
        self._log_level = log_level
        self._logger.setLevel(self._log_level)

        self._logger.debug("Initializing the TFC API class...")

        self._instance_url = url
        self._token = api_token
        self._current_org = None
        self._verify = verify

        self.admin_module_sharing: TFCAdminModuleSharing = None
        self.admin_orgs: TFCAdminOrgs = None
        self.admin_runs: TFCAdminRuns = None
        self.admin_settings: TFCAdminSettings = None
        self.admin_terraform_versions: TFCAdminTerraformVersions = None
        self.admin_users: TFCAdminUsers = None
        self.admin_workspaces: TFCAdminWorkspaces = None
        self.audit_trails: TFCAuditTrails = None
        self.invoices: TFCInvoices = None
        self.ip_ranges: TFCIPRanges = None
        self.feature_sets: TFCFeatureSets = None
        self.orgs: TFCOrgs = None
        self.account: TFCAccount = None
        self.applies: TFCApplies = None
        self.agents: TFCAgents = None
        self.agent_tokens: TFCAgentTokens = None
        self.config_versions: TFCConfigVersions = None
        self.cost_estimates: TFCCostEstimates = None
        self.oauth_clients: TFCOAuthClients = None
        self.oauth_tokens: TFCOAuthTokens = None
        self.org_memberships: TFCOrgMemberships = None
        self.org_tokens: TFCOrgTokens = None
        self.plans: TFCPlans = None
        self.plan_exports: TFCPlanExports = None
        self.policies: TFCPolicies = None
        self.policy_checks: TFCPolicyChecks = None
        self.policy_sets: TFCPolicySets = None
        self.policy_set_params: TFCPolicySetParams = None
        self.notification_configs: TFCNotificationConfigurations = None
        self.registry_modules: TFCRegistryModules = None
        self.run_triggers: TFCRunTriggers = None
        self.runs: TFCRuns = None
        self.state_versions: TFCStateVersions = None
        self.state_version_outputs: TFCStateVersionOutputs = None
        self.ssh_keys: TFCSSHKeys = None
        self.teams: TFCTeams = None
        self.team_access: TFCTeamAccess = None
        self.team_memberships: TFCTeamMemberships = None
        self.team_tokens: TFCTeamTokens = None
        self.users: TFCUsers = None
        self.user_tokens: TFCUserTokens = None
        self.vars: TFCVars = None
        self.workspace_vars: TFCWorkspaceVars = None
        self.workspaces: TFCWorkspaces = None

        self._token = None
        self._headers = None

        self._logger.debug("Retrieving TFC API well known paths..")
        self._well_known_paths = self.well_known_paths()
        self._logger.debug("TFC API well known paths retrieved.")

        self.set_token(api_token)
        self._initialize_endpoints()

    def _get(self, url):
        """
        Simplified HTTP GET function for usage only with this API module.
        """
        results = None
        req = requests.get(url, headers=self._headers, verify=self._verify)

        if req.status_code == HTTP_OK:
            results = json.loads(req.content)
            self._logger.debug(f"GET to {url} successful")
        else:
            err = json.loads(req.content.decode("utf-8"))
            self._logger.debug(err)

        return results

    def _initialize_endpoints(self):
        self._logger.debug("Initializing endpoints that don't require an org to be set...")
        # Loop through all the endpoints that don't require an org and initialize them
        for ep_name in self._class_for_attr_dict["org-not-required"]:
            class_for_attr = self._class_for_attr_dict["org-not-required"][ep_name]
            initialized_endpoint_class = class_for_attr(
                self._instance_url,
                None,
                self._headers,
                self._well_known_paths,
                self._verify,
                self._log_level)
            setattr(self, ep_name, initialized_endpoint_class)

        self._logger.debug("Initialized endpoints that don't require an org to be set.")
        # Loop through all the endpoints that do require an org and initialize them
        for ep_name in self._class_for_attr_dict["org-required"]:
            setattr(self, ep_name, None)

        self._logger.debug("TFC API class initialized.")

    def get_url(self):
        """
        Allows for the user to retrieve the URL being hit from the API object.
        """
        return self._instance_url

    def set_org(self, org_name):
        """
        Sets the organization to use for org specific endpoints.
        This method must be called for any non-admin endpoint to work.
        """

        self._logger.debug("Initializing endpoints that do require an org to be set...")

        # Update the current org attribute
        self._current_org = org_name

        for ep_name in self._class_for_attr_dict["org-required"]:
            class_for_attr = self._class_for_attr_dict["org-required"][ep_name]
            initialized_endpoint_class = class_for_attr(
                self._instance_url,
                self._current_org,
                self._headers,
                self._well_known_paths,
                self._verify,
                self._log_level)

            setattr(self, ep_name, initialized_endpoint_class)

        self._logger.debug("Initialized endpoints that do require an org to be set.")

    def get_org(self):
        """
        Allows for the user to retrieve the current org from the API object.
        """
        return self._current_org

    def get_entitlements(self):
        """
        Allows for the user to retrieve the entitlements to the API for the current org.
        """
        entitlements = None

        if self.is_terraform_cloud():
            try:
                entitlements = self.orgs.entitlements(self._current_org)["data"]["attributes"]
            except TFCHTTPNotFound:
                self._logger.debug("Entitlements API endpoint not found. No entitlements recorded.")
        else:
            self._logger.debug("Not Terraform Cloud, so entitlements API is not supported.")

        return entitlements

    def set_token(self, token):
        """
        Allows for the user to change the token they are using on the fly if
        they need to change tokens.
        """
        self._token = token
        self._headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/vnd.api+json"
        }
        self._initialize_endpoints()

    def get_token(self):
        """
        Allows for the user to retrieve the token from the API object.
        """
        return self._token

    def is_terraform_cloud(self):
        return TFC_SAAS_HOSTNAME in self._instance_url


    def well_known_paths(self):
        """
        Retrieve all the well known paths from the Terraform Cloud installation.
        """
        url = f"{self._instance_url}/.well-known/terraform.json"
        return self._get(url)



"""
Module for Terraform Cloud API Endpoint: IP Ranges.
"""

from .endpoint import TFCEndpoint

class TFCIPRanges(TFCEndpoint):
    """
    `IP Ranges API Docs \
        <https://www.terraform.io/docs/cloud/api/ip-ranges.html>`_
    """

    def __init__(self, instance_url, org_name, headers, well_known_paths, verify, log_level):
        super().__init__(instance_url, org_name, headers, well_known_paths, verify, log_level)
        self._endpoint_base_url = f"{self._meta_base_url}/ip-ranges"

    def required_entitlements(self):
        return []

    def terraform_cloud_only(self):
        return True

    def terraform_enterprise_only(self):
        return False

    def list(self):
        """
        ``GET /meta/ip-ranges``

        `IP Ranges List API Doc Reference \
            <https://www.terraform.io/docs/cloud/api/ip-ranges.html#get-ip-ranges>`_
        """
        return self._list(self._endpoint_base_url)
